from __future__ import annotations

import math
import re
import struct
from dataclasses import dataclass

import pwndbg
import pwndbg.aglib.disasm.disassembly
import pwndbg.aglib.kernel
import pwndbg.aglib.kernel.symbol
import pwndbg.aglib.kernel.vmmap
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.lib.cache
import pwndbg.lib.memory
import pwndbg.lib.regs
import pwndbg.search
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.lib.memory import Page
from pwndbg.lib.regs import BitFlags

# don't return None but rather an invalid value for address markers
# this way arithmetic ops do not panic if physmap is not found
INVALID_ADDR = 1 << 64


@pwndbg.lib.cache.cache_until("stop")
def first_kernel_page_start() -> int:
    for page in pwndbg.aglib.kernel.vmmap.kernel_vmmap_pages():
        if page.start and pwndbg.aglib.memory.is_kernel(page.start):
            return page.start
    return INVALID_ADDR


@dataclass
class PageTableLevel:
    name: str | None = None
    entry: int | None = None
    virt: int | None = None  # within physmap
    phys: int | None = None  # physcal address
    idx: int | None = None
    level: int | None = None


@dataclass
class PagewalkResult:
    virt: int | None = None  # within physmap
    phys: int | None = None  # physcal address
    entry: int | None = None
    levels: tuple[PageTableLevel, ...] = ()


class PageTableScan:
    def __init__(self, pi: ArchPagingInfo) -> None:
        # from ArchPagingInfo:
        self.paging_level = pi.paging_level
        self.PAGE_ENTRY_MASK = pi.PAGE_ENTRY_MASK
        self.PAGE_INDEX_LEN = pi.PAGE_INDEX_LEN
        self.PAGE_INDEX_MASK = pi.PAGE_INDEX_MASK
        self.page_shift = pi.page_shift
        self.should_stop_pagewalk = pi.should_stop_pagewalk
        # for scanning
        self.pagesz = pi.page_size
        self.ptrsize = pwndbg.aglib.arch.ptrsize
        self.inf = pwndbg.dbg.selected_inferior()
        self.fmt = "<" + ("Q" if self.ptrsize == 8 else "I") * (self.pagesz // self.ptrsize)
        self.cache: dict[tuple[int, int], list[tuple[int, int, int]]] = {}
        self.entry_cache: dict[int, tuple[int]] = {}
        self.arch = pwndbg.aglib.arch.name

    def scan(self, entry: int, is_kernel: bool = False) -> list[Page]:
        """
        this needs to be EXTREMELY optimized as it is used to display context
        making as few functions calls or memory reads as possible
        avoid unnecessary python pointer deferences or repetative computations whenever possible
        when benchmarked on the same linux kernels, on average:
        - gdb-pt-dump takes ~0.153 for x64 and 5.572 seconds for aarch64
        - this implementation takes less than 0.065 seconds to complete for x64 and 0.491 seconds for aarch64
        --> around 45-65% of the time is used to read qemu system memory depending on arch and kernel
            (the theoratical limit would be that all time consumed is used for reading memory)
        --> 2.35x speed up for x64 and more than 10x speed up for aarch64
        """
        entry &= self.PAGE_ENTRY_MASK
        if (entry, self.paging_level) not in self.cache:
            self._scan(entry, self.paging_level)
        result = []
        curr = None
        kernel_prefix_shift = self.paging_level * self.PAGE_INDEX_LEN + self.page_shift
        # assumes the elements in self.cache[*] are sorted and non-overlapping
        for offset, size, flags in self.cache[(entry, self.paging_level)]:
            if self.arch == "x86-64":
                is_kernel = offset >= (1 << 47)
            if is_kernel:
                nbits = self.ptrsize * 8 - kernel_prefix_shift
                offset += ((1 << nbits) - 1) << kernel_prefix_shift
            if curr and offset == curr.end and flags == curr.flags:
                # merge contiguous chunks
                curr.memsz = max(curr.memsz, offset + size - curr.start)
            else:
                if curr:
                    result.append(curr)
                curr = Page(offset, size, flags, 0, self.ptrsize)
        if curr:
            result.append(curr)
        return result

    def _scan(self, addr: int, level: int) -> None:
        pagesz = self.pagesz
        orig = addr
        self.entry_cache[addr] = struct.unpack(self.fmt, self.inf.read_memory(addr, self.pagesz))
        entries = self.entry_cache[addr]
        ranges: list[tuple[int, int, int]] = []
        append = ranges.append
        # the range currently being merged, curr_off == None means there is no current range being merged
        curr_off = None
        curr_sz = curr_flags = 0
        # len(entries) == self.pagesz // self.ptrsize, try not to do division here
        size = pagesz * (len(entries) ** (level - 1))
        # per entry offset relative to the start of the current pagetable,
        # each entry represents `size` bytes of memory, therefore, offset += size for each entry
        offset = 0
        # TODO: prev is used to avoid clustering the vmmap output with espfix ranges (happens for x86-64)
        # consecutive identical pt entries will be clapsed into one. None means not x86-64
        # I believe this is what gdb-pt-dump does as this gives identical output
        prev = 0 if self.arch == "x86-64" else None
        for entry in entries:
            if prev and prev == entry:
                # be aware to not use continue
                pass
            elif entry == 0:
                if curr_off is not None:
                    append((curr_off, curr_sz, curr_flags))
                    curr_off = None
            elif level == 1 or self.should_stop_pagewalk(entry):
                flags = 0
                match self.arch:
                    case "x86-64":
                        flags = Page.R_OK if entry & 1 != 0 else 0
                        flags |= Page.W_OK if entry & (1 << 1) else 0
                        flags |= Page.X_OK if entry & (1 << 63) == 0 else 0
                    case "aarch64":
                        flags = Page.R_OK if entry & 1 != 0 else 0
                        flags |= Page.X_OK if (entry >> 53) & 3 != 3 else 0
                        ap = (entry >> 6) & 3
                        flags |= Page.W_OK if ap == 1 or ap == 0 else 0
                if flags & Page.R_OK:  # only append present pages, read bit indicates presence
                    if curr_off is not None:
                        if flags == curr_flags:
                            curr_sz += size
                        else:
                            append((curr_off, curr_sz, curr_flags))
                            curr_off = None
                    if curr_off is None:
                        curr_off, curr_sz, curr_flags = offset, size, flags
            else:
                addr = entry & self.PAGE_ENTRY_MASK
                key = (addr, level - 1)
                if key not in self.cache:
                    # only call when should keep scanning the page tree
                    # we need to reduce this recursive call as much as possible
                    # each time the level is decremented, garanteed to terminate
                    self._scan(addr, level - 1)
                arr = self.cache[key]
                # The following if-blocks are purely for optimization purposes
                # coalesce as much as we can
                left, n = 0, len(arr)
                right = n - 1
                if n > 0:  # merge the first page chunk range if needed
                    if curr_off is not None:
                        roff, rsz, rflags = arr[0]
                        # is the first non-zero entry actually the first (0 th) entry?
                        if rflags == curr_flags and roff == 0:
                            curr_sz += rsz
                            left += 1
                if curr_off is not None and (n > 1 or (n == 1 and left == 0)):
                    append((curr_off, curr_sz, curr_flags))
                    curr_off = None
                if n > 1:  # (prepare to) merge the last page chunk range if needed
                    roff, rsz, rflags = arr[n - 1]
                    # is the last non-zero entry actually the last (e.g. 511 th) entry?
                    if roff + rsz == size:
                        curr_off, curr_sz, curr_flags = offset + roff, rsz, rflags
                        right -= 1
                for roff, rsz, rflags in arr[left : right + 1]:
                    append((offset + roff, rsz, rflags))
            offset += size
            if prev is not None:
                prev = entry
        if curr_off is not None:
            append((curr_off, curr_sz, curr_flags))
        self.cache[(orig, level)] = ranges

    def walk(self, target: int, entry: int) -> PagewalkResult:
        page_shift = self.page_shift
        levels = [PageTableLevel() for _ in range(self.paging_level)]
        resolved = offset_mask = None
        for i in range(self.paging_level - 1, -1, -1):
            resolved = None
            shift = page_shift + self.PAGE_INDEX_LEN * i
            idx = (target >> shift) & self.PAGE_INDEX_MASK
            addr = entry & self.PAGE_ENTRY_MASK
            if addr not in self.entry_cache:
                self.entry_cache[addr] = struct.unpack(
                    self.fmt, self.inf.read_memory(addr, self.pagesz)
                )
            entry = self.entry_cache[addr][idx]
            if not entry:
                break
            levels[i].phys = addr
            levels[i].idx = idx
            levels[i].entry = entry
            levels[i].level = i + 1
            offset_mask = (1 << shift) - 1
            resolved = (entry & self.PAGE_ENTRY_MASK, offset_mask)
            if self.should_stop_pagewalk(entry):
                break
        result = PagewalkResult()
        if resolved and offset_mask is not None:
            addr, offset_mask = resolved
            result.phys = addr + (target & offset_mask)
            result.entry = entry
        result.levels = tuple(levels)
        return result


class ArchPagingInfo:
    USERLAND = "userland"
    KERNELLAND = "kernel [.text]"
    KERNELRO = "kernel [.rodata]"
    KERNELBSS = "kernel [.bss]"
    KERNELDRIVER = "kernel [.driver .bpf]"
    ESPSTACK = "espfix"
    PHYSMAP = "physmap"
    VMALLOC = "vmalloc"
    VMEMMAP = "vmemmap"

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def STRUCT_PAGE_SIZE(self) -> int:
        a = pwndbg.aglib.typeinfo.load("struct page")
        if a is None:
            # true with the most common set of configurations
            # this struct should always present if a vmlinux is added
            return 0x40
        # needs to be rounded up to a power of 2 (consider the layout of vmemmap)
        return 1 << math.ceil(math.log2(a.sizeof))

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def STRUCT_PAGE_SHIFT(self) -> int:
        return int(math.log2(self.STRUCT_PAGE_SIZE))

    @property
    def physmap(self) -> int:
        raise NotImplementedError()

    @property
    def vmalloc(self) -> int:
        raise NotImplementedError()

    @property
    def vmemmap(self) -> int:
        raise NotImplementedError()

    def slab_to_virt(self, slab: int) -> int:
        raise NotImplementedError()

    def virt_to_slab(self, virt: int) -> int:
        raise NotImplementedError()

    @property
    def slab_virtual(self) -> int:
        raise NotImplementedError()

    @property
    def kbase(self) -> int | None:
        raise NotImplementedError()

    @property
    def page_shift(self) -> int:
        raise NotImplementedError()

    @property
    def page_size(self) -> int:
        return 1 << self.page_shift

    @property
    def paging_level(self) -> int:
        raise NotImplementedError()

    def adjust(self, name: str) -> str:
        raise NotImplementedError()

    @pwndbg.lib.cache.cache_until("stop")
    def markers(self) -> tuple[tuple[str | None, int | None], ...]:
        raise NotImplementedError()

    def handle_kernel_pages(self, pages: tuple[Page, ...]) -> None:
        # this is arch dependent
        raise NotImplementedError()

    def _kbase(self, address: int | None) -> int | None:
        if address is None:
            return None
        for mapping in pwndbg.aglib.kernel.vmmap.kernel_vmmap_pages():
            # should be page aligned -- either from pt-dump or info mem

            # only search in kernel mappings:
            # https://www.kernel.org/doc/html/v5.3/arm64/memory.html
            if not pwndbg.aglib.memory.is_kernel(mapping.vaddr):
                continue

            if address in mapping:
                return mapping.vaddr

        return None

    def pagewalk(self, target: int, entry: int | None, virt: bool = True) -> PagewalkResult:
        raise NotImplementedError()

    def pagescan(self, entry: int | None = None) -> list[Page]:
        raise NotImplementedError()

    @property
    def PAGE_ENTRY_MASK(self) -> int:
        return ~(self.page_size - 1) & ((1 << self.va_bits) - 1)

    @property
    def PAGE_INDEX_LEN(self) -> int:
        return self.page_shift - math.ceil(math.log2(pwndbg.aglib.arch.ptrsize))

    @property
    def PAGE_INDEX_MASK(self) -> int:
        return (1 << (self.PAGE_INDEX_LEN)) - 1

    @pwndbg.lib.cache.cache_until("stop")
    def pagetablescan(self, entry: int) -> PageTableScan | None:
        return PageTableScan(self)

    def switch_to_phymem_mode(self) -> tuple[str, bool]:
        oldval = pwndbg.dbg.selected_inferior().send_remote("qqemu.PhyMemMode").decode()
        pwndbg.dbg.selected_inferior().send_remote("Qqemu.PhyMemMode:1")
        # only two possible return values: https://qemu-project.gitlab.io/qemu/system/gdb.html
        success = pwndbg.dbg.selected_inferior().send_remote("qqemu.PhyMemMode") == b"1"
        return oldval, success

    def _pagewalk(self, target: int, entry: int, virt: bool) -> PagewalkResult:
        scan = self.pagetablescan(entry)
        oldval, success = self.switch_to_phymem_mode()
        if not success or not scan:
            return PagewalkResult()
        result = PagewalkResult()
        try:
            result = scan.walk(target, entry)
            if not virt:
                return result
            levels = result.levels
            for level in levels:
                if level.phys is None or level.level is None:
                    continue
                level.virt = level.phys + self.physmap - self.phys_offset
                level.name = self.pagetable_level_names[level.level]
            if result.phys is not None:
                result.virt = result.phys + self.physmap - self.phys_offset
        except Exception:
            pass
        finally:  # so that the PhyMemMode value is always restored
            pwndbg.dbg.selected_inferior().send_remote(f"Qqemu.PhyMemMode:{oldval}")
        return result

    def _pagescan(self, entry: int, is_kernel: bool = False) -> list[Page]:
        scan = self.pagetablescan(entry)
        oldval, success = self.switch_to_phymem_mode()
        if not success or not scan:
            return []
        result = []
        try:
            result = scan.scan(entry, is_kernel)
        except Exception:
            pass
        finally:  # so that the PhyMemMode value is always restored
            pwndbg.dbg.selected_inferior().send_remote(f"Qqemu.PhyMemMode:{oldval}")
        return result

    def bitflags(self, level: PageTableLevel) -> BitFlags:
        raise NotImplementedError()

    def should_stop_pagewalk(self, entry: int) -> bool:
        raise NotImplementedError()

    @property
    def phys_offset(self) -> int:
        return 0

    @property
    def va_bits(self) -> int:
        raise NotImplementedError()

    @property
    def pagetable_level_names(self) -> tuple[str, ...]:
        raise NotImplementedError()


class x86_64PagingInfo(ArchPagingInfo):
    def __init__(self) -> None:
        self.P4D_SHIFT = 39

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def pagetable_level_names(self) -> tuple[str, ...]:
        # https://blog.zolutal.io/understanding-paging/
        match self.paging_level:
            case 4:
                return ("Page", "PT", "PMD", "PUD", "PGD")
            case 5:
                return ("Page", "PT", "PMD", "P4D", "PUD", "PGD")
        return ()

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def va_bits(self) -> int:
        return 48 if self.paging_level == 4 else 51

    @pwndbg.lib.cache.cache_until("stop")
    def get_vmalloc_vmemmap_bases(self) -> tuple[int | None, int | None]:
        result = None
        try:
            target = self.physmap.to_bytes(8, byteorder="little")
            mapping = pwndbg.aglib.kernel.first_kernel_ro_page()
            if not mapping:
                return None, None
            result = next(
                pwndbg.search.search(target, mappings=[mapping], aligned=pwndbg.aglib.arch.ptrsize),
                None,
            )
        except Exception as e:
            print(e)
            pass
        vmemmap, vmalloc = None, None
        if result is not None:
            vmemmap = pwndbg.aglib.memory.u64(result - 0x10)
            vmalloc = pwndbg.aglib.memory.u64(result - 0x8)
        return vmalloc, vmemmap

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def physmap(self) -> int:
        result: int | None = pwndbg.aglib.kernel.symbol.try_usymbol("page_offset_base")
        if not result:
            result = first_kernel_page_start()
        return result

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def kbase(self) -> int | None:
        idt_entries = pwndbg.aglib.kernel.get_idt_entries()
        if len(idt_entries) == 0:
            return None
        return self._kbase(idt_entries[0].offset)

    @property
    def page_shift(self) -> int:
        return 12

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def vmalloc(self) -> int:
        result: int | None = pwndbg.aglib.kernel.symbol.try_usymbol("vmalloc_base")
        if result is not None:
            return result
        result, _ = self.get_vmalloc_vmemmap_bases()
        if result is not None:
            return result
        # resort to default
        return 0xFF91000000000000 if self.paging_level == 5 else 0xFFFFC88000000000

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def vmemmap(self) -> int:
        result: int | None = pwndbg.aglib.kernel.symbol.try_usymbol("vmemmap_base")
        if result is not None:
            return result
        _, result = self.get_vmalloc_vmemmap_bases()
        if result is not None:
            return result
        # resort to default
        return 0xFFD4000000000000 if self.paging_level == 5 else 0xFFFFEA0000000000

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def slab_virtual(self) -> int:
        return pwndbg.aglib.arch.unsigned(-3, self.P4D_SHIFT)

    @property
    def SLAB_DATA_BASE_ADDR(self):
        STRUCT_SLAB_SIZE = 32 * pwndbg.aglib.arch.ptrsize
        SLAB_VPAGES = (1 << self.P4D_SHIFT) // self.page_size
        SLAB_META_SIZE = pwndbg.lib.memory.round_up(STRUCT_SLAB_SIZE * SLAB_VPAGES, self.page_size)
        return self.slab_virtual + SLAB_META_SIZE

    def slab_to_virt(self, slab: int) -> int:
        a = pwndbg.aglib.typeinfo.load("struct slab")
        # default is true for mitigation-6.12
        # TODO: use heuristics to derive the size of the struct
        slab_size = (14 * pwndbg.aglib.arch.ptrsize) if not a else a.sizeof
        idx = (slab - self.slab_virtual) // slab_size
        return self.SLAB_DATA_BASE_ADDR + self.page_size * idx

    def virt_to_slab(self, virt: int) -> int:
        a = pwndbg.aglib.typeinfo.load("struct slab")
        # default is true for mitigation-6.12
        slab_size = (14 * pwndbg.aglib.arch.ptrsize) if not a else a.sizeof
        slab = self.slab_virtual + (virt - self.SLAB_DATA_BASE_ADDR) // self.page_size * slab_size
        slab = pwndbg.aglib.memory.get_typed_pointer("struct slab", slab)
        return int(slab["compound_slab_head"])

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def paging_level(self) -> int:
        cr4 = pwndbg.aglib.regs.read_reg("cr4")
        return 4 if cr4 is None or (cr4 & (1 << 12)) == 0 else 5

    @pwndbg.lib.cache.cache_until("stop")
    def markers(self) -> tuple[tuple[str | None, int | None], ...]:
        # https://www.kernel.org/doc/Documentation/x86/x86_64/mm.txt
        return (
            (self.USERLAND, 0),
            (None, 0x8000000000000000),
            ("ldt remap", 0xFFFF880000000000 if self.paging_level == 4 else 0xFF10000000000000),
            (self.PHYSMAP, self.physmap),
            (self.VMALLOC, self.vmalloc),
            (self.VMEMMAP, self.vmemmap),
            ("cpu entry", 0xFFFFFE0000000000),
            ("slab virtual", self.slab_virtual),
            (None, self.slab_virtual + (1 << self.P4D_SHIFT)),
            (self.ESPSTACK, 0xFFFFFF0000000000),
            ("EFI", 0xFFFFFFEF00000000),
            (self.KERNELLAND, self.kbase),
            ("fixmap", 0xFFFFFFFFFF000000),
            ("legacy abi", 0xFFFFFFFFFF600000),
            (None, 0xFFFFFFFFFFFFFFFF),
        )

    def adjust(self, name: str) -> str:
        name = name.lower()
        if "low kernel" in name:
            return self.PHYSMAP
        if "high kernel" in name:
            return self.KERNELLAND
        if self.VMALLOC in name:
            return self.VMALLOC
        if self.VMEMMAP in name:
            return self.VMEMMAP
        if " area" in name:
            return name[:-5]
        return name

    def handle_kernel_pages(self, pages: tuple[Page, ...]) -> None:
        kernel_idx = None
        kbase = self.kbase
        stack = pwndbg.aglib.regs.read_reg(pwndbg.aglib.regs.stack)
        for i, page in enumerate(pages):
            if kernel_idx is None and kbase is not None and kbase in page:
                kernel_idx = i
                break
        if kernel_idx is None:
            return
        has_loadable_driver = False
        for i in range(kernel_idx, len(pages)):
            page = pages[i]
            if page.objfile != self.KERNELLAND:
                break
            if page.start == kbase:
                continue
            # the first executable page after kernel text is the start of bpf/loadable driver
            if has_loadable_driver or page.execute:
                page.objfile = self.KERNELDRIVER
                has_loadable_driver = True
                continue
            if not page.execute:
                if page.write:
                    page.objfile = self.KERNELBSS
                else:
                    page.objfile = self.KERNELRO
        if not stack:
            return
        for i, page in enumerate(pages):
            if stack in page:
                page.objfile += " [stack]"

    def pagewalk(self, target: int, entry: int | None, virt: bool = True) -> PagewalkResult:
        if entry is None:
            entry = pwndbg.aglib.regs.read_reg("cr3")
        if entry is None:
            return PagewalkResult()
        return self._pagewalk(target, entry, virt)

    def pagescan(self, entry: int | None = None) -> list[Page]:
        if entry is None:
            entry = pwndbg.aglib.regs.read_reg("cr3")
        if entry is None:
            return []
        return self._pagescan(entry)

    def bitflags(self, level: PageTableLevel) -> BitFlags:
        return BitFlags([("NX", 63), ("PS", 7), ("A", 5), ("U", 2), ("W", 1), ("P", 0)])

    def should_stop_pagewalk(self, entry: int) -> bool:
        return entry & (1 << 7) > 0


class Aarch64PagingInfo(ArchPagingInfo):
    def __init__(self) -> None:
        self.VMEMMAP_START = self.VMEMMAP_SIZE = self.PAGE_OFFSET = None

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def pagetable_level_names(self) -> tuple[str, ...]:
        match self.paging_level:
            case 4:
                return ("Page", "L3", "L2", "L1", "L0")
            case 3:
                return ("Page", "L3", "L2", "L1")
            case 2:
                return ("Page", "L3", "L2")
        return ()

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def tcr_el1(self) -> BitFlags:
        tcr = pwndbg.lib.regs.aarch64_tcr_flags
        tcr.value = pwndbg.aglib.regs.read_reg("TCR_EL1") or 0
        return tcr

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def va_bits(self) -> int:
        id_aa64mmfr2_el1 = pwndbg.lib.regs.aarch64_mmfr_flags
        id_aa64mmfr2_el1.value = pwndbg.aglib.regs.read_reg("ID_AA64MMFR2_EL1") or 0
        feat_lva = id_aa64mmfr2_el1.value is not None and id_aa64mmfr2_el1["VARange"] == 0b0001
        va_bits: int = 64 - self.tcr_el1["T1SZ"]  # this is prob only `vabits_actual`
        self.PAGE_OFFSET = self._PAGE_OFFSET(va_bits)  # physmap base address without KASLR
        if feat_lva:
            va_bits = min(52, va_bits)
        return va_bits

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def va_bits_min(self) -> int:
        return 48 if self.va_bits > 48 else self.va_bits

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def vmalloc(self) -> int:
        # also includes KASAN and kernel module regions
        return self._PAGE_END(self.va_bits_min)

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def physmap(self) -> int:
        return first_kernel_page_start()

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def kbase(self) -> int | None:
        return self._kbase(pwndbg.aglib.regs.read_reg("vbar"))

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def kversion(self) -> tuple[int, ...] | None:
        return pwndbg.aglib.kernel.krelease()

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def module_start(self) -> int | None:
        if self.kbase is None:
            return None
        res = None
        for page in pwndbg.aglib.kernel.vmmap.kernel_vmmap_pages()[::-1]:
            if page.start >= self.kbase:
                continue
            if page.start < self.vmalloc:
                break
            if page.execute:
                res = page.start
                break
        return res

    def _PAGE_OFFSET(self, va: int) -> int:  # aka PAGE_START
        return pwndbg.aglib.arch.unsigned(-(1 << va))

    def _PAGE_END(self, va: int) -> int:
        return pwndbg.aglib.arch.unsigned(-(1 << (va - 1)))

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def vmemmap(self) -> int:
        _ = self.va_bits_min
        if self.kversion is None or self.PAGE_OFFSET is None:
            return INVALID_ADDR
        vmemmap_shift = self.page_shift - self.STRUCT_PAGE_SHIFT
        # self.PAGE_OFFSET is set by self.va_bits(_min) so must exist
        if self.kversion < (5, 4):
            self.VMEMMAP_SIZE = 1 << (self.va_bits - self.page_shift - 1 + self.STRUCT_PAGE_SHIFT)
            self.VMEMMAP_START = self.PAGE_OFFSET - self.VMEMMAP_SIZE
        elif self.kversion < (5, 11):
            self.VMEMMAP_SIZE = (
                self._PAGE_END(self.va_bits_min) - self.PAGE_OFFSET
            ) >> vmemmap_shift
            self.VMEMMAP_START = pwndbg.aglib.arch.unsigned(-self.VMEMMAP_SIZE - 0x00200000)
        elif self.kversion < (6, 9):
            self.VMEMMAP_SIZE = (
                self._PAGE_END(self.va_bits_min) - self.PAGE_OFFSET
            ) >> vmemmap_shift
            self.VMEMMAP_START = self._PAGE_OFFSET(self.va_bits - vmemmap_shift)
        else:
            VMEMMAP_RANGE = self._PAGE_END(self.va_bits_min) - self.PAGE_OFFSET
            self.VMEMMAP_SIZE = (VMEMMAP_RANGE >> self.page_shift) * self.STRUCT_PAGE_SIZE
            self.VMEMMAP_START = pwndbg.aglib.arch.unsigned(-0x40000000 - self.VMEMMAP_SIZE)

        # obtained through debugging -- kaslr offset of physmap determines the offset of vmemmap
        vmemmap_kaslr = (self.physmap - self.PAGE_OFFSET - self.phys_offset) >> vmemmap_shift
        return self.VMEMMAP_START + vmemmap_kaslr

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def pci(self) -> int | None:
        if self.kversion is None or self.VMEMMAP_START is None or self.VMEMMAP_SIZE is None:
            return None
        self.pci_end = INVALID_ADDR
        if self.kversion >= (6, 9):
            pci = self.VMEMMAP_START + self.VMEMMAP_SIZE + 0x00800000  # 8M
            self.pci_end = pci + 0x01000000  # 16M
            return pci
        if self.kversion >= (5, 11):
            self.pci_end = self.VMEMMAP_START - 0x00800000  # 8M
        else:
            self.pci_end = self.VMEMMAP_START - 0x00200000  # 2M
        return self.pci_end - 0x01000000  # 16M

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def fixmap(self) -> int:
        if self.kversion is None:
            return INVALID_ADDR
        if self.pci and self.kversion < (5, 11):
            FIXADDR_TOP = self.pci - 0x00200000  # 2M
        elif self.VMEMMAP_START and self.kversion < (6, 9):
            FIXADDR_TOP = self.VMEMMAP_START - 0x02000000  # 32M
        else:
            FIXADDR_TOP = pwndbg.aglib.arch.unsigned(-0x00800000)
        # https://elixir.bootlin.com/linux/v6.16.5/source/arch/arm64/include/asm/fixmap.h#L102
        # 0x1000 is an upper estimate
        FIXADDR_SIZE = 0x1000 << self.page_shift
        return FIXADDR_TOP - FIXADDR_SIZE

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def ksize(self) -> int:
        start = pwndbg.aglib.symbol.lookup_symbol_addr("_text")
        if start is None:
            start = self.kbase
        end = pwndbg.aglib.symbol.lookup_symbol_addr("_end")
        if start is not None and end is not None:
            return end - start
        # fallback
        return 100 << 21  # 100M

    @property
    @pwndbg.lib.cache.cache_until("objfile")
    def page_shift_heuristic(self) -> int:
        default_val = 12
        sym = pwndbg.aglib.symbol.lookup_symbol_addr("copy_page_to_iter")
        if sym is not None:
            pattern = re.compile(r".*(0x1000|0x10000|0x4000)")
            addr = int(sym)
            # sanity check.
            if not pwndbg.aglib.memory.peek(addr):
                return default_val

            for _ in range(50):
                # It is **crucial** that we don't enhance, because enhancing will run vmmap
                # which will end up running this code, causing infinite recursion.
                # https://github.com/pwndbg/pwndbg/actions/runs/20342890859/job/58446963784?pr=3512
                instr: PwndbgInstruction = pwndbg.aglib.disasm.disassembly.get_one_instruction(
                    addr, enhance=False
                )

                if instr.mnemonic == "MOV" and (result := pattern.search(instr.op_str)) is not None:
                    return int(math.log2(int(result.group(1), 16)))

                addr = instr.next

        return default_val

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def page_shift(self) -> int:
        match self.tcr_el1["TG1"]:
            case 0b01:
                return 14
            case 0b10:
                return 12
            case 0b11:
                return 16
        return self.page_shift_heuristic

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def page_shift_user(self) -> int:
        match self.tcr_el1["TG0"]:
            case 0b00:
                return 12
            case 0b01:
                return 16
            case 0b10:
                return 14
        return self.page_shift_heuristic

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def paging_level(self) -> int:
        # https://www.kernel.org/doc/html/v5.3/arm64/memory.html
        if self.page_shift == 16:
            return 2
        # in some cases, not all addressing bits are used
        return (self.va_bits - self.page_shift + (self.page_shift - 4)) // (self.page_shift - 3)

    @pwndbg.lib.cache.cache_until("stop")
    def markers(self) -> tuple[tuple[str | None, int | None], ...]:
        address_markers = pwndbg.aglib.symbol.lookup_symbol_addr("address_markers")
        if address_markers is not None:
            sections: list[tuple[str | None, int | None]] = [(self.USERLAND, 0)]
            value = 0
            name = None
            for i in range(20):
                value = pwndbg.aglib.memory.u64(address_markers + i * 0x10)
                name_ptr = pwndbg.aglib.memory.u64(address_markers + i * 0x10 + 8)
                name = None
                if name_ptr > 0:
                    name = pwndbg.aglib.memory.string(name_ptr).decode()
                    name = self.adjust(name)
                if value > 0:
                    sections.append((name, value))
                if value == pwndbg.aglib.arch.unsigned(-1):
                    break
            return tuple(sections)
        vmalloc_end = None
        if self.vmemmap and self.pci and self.fixmap:
            vmalloc_end = min(self.vmemmap, self.pci, self.fixmap)
        if self.VMEMMAP_START is None or self.VMEMMAP_SIZE is None or self.PAGE_OFFSET is None:
            return ()
        return (
            (self.USERLAND, 0),
            (None, self.PAGE_OFFSET),
            (self.PHYSMAP, self.physmap),
            (None, self.vmalloc),
            (self.VMALLOC, self.vmalloc),
            (None, vmalloc_end),
            (self.VMEMMAP, self.vmemmap),
            (None, self.VMEMMAP_START + self.VMEMMAP_SIZE),
            ("pci", self.pci),
            (None, self.pci_end),
            ("fixmap", self.fixmap),
            (None, pwndbg.aglib.arch.unsigned(-1)),
        )

    def adjust(self, name: str) -> str:
        name = name.lower()
        if "end" in name:
            return ""
        if "linear" in name:
            return self.PHYSMAP
        if "modules" in name:
            return self.KERNELDRIVER
        if self.VMEMMAP in name:
            return self.VMEMMAP
        if self.VMALLOC in name:
            return self.VMALLOC
        return " ".join(name.strip().split()[:-1])

    def handle_kernel_pages(self, pages: tuple[Page, ...]) -> None:
        if self.kbase is None:
            return
        stack = pwndbg.aglib.regs.read_reg(pwndbg.aglib.regs.stack)
        for i in range(len(pages)):
            page = pages[i]
            if self.module_start and self.module_start <= page.start < self.kbase:
                page.objfile = self.KERNELDRIVER
            elif self.kbase <= page.start < self.kbase + self.ksize:
                page.objfile = self.KERNELLAND
                if not page.execute:
                    if page.write:
                        page.objfile = self.KERNELBSS
                    else:
                        page.objfile = self.KERNELRO
            if stack and stack in page:
                page.objfile += " [stack]"

    @property
    @pwndbg.lib.cache.cache_until("start")
    def phys_offset(self) -> int:
        found_system = False
        try:
            for line in pwndbg.dbg.selected_inferior().send_monitor("info mtree -f").splitlines():
                line = line.strip()
                if "Root memory region: system" in line:
                    found_system = True
                if found_system:
                    split = line.split("-")
                    if "ram" in line and len(split) > 1:
                        return int(split[0], 16)
        except Exception:
            pass
        return 0x40000000  # default

    def pagewalk(self, target: int, entry: int | None, virt: bool = True) -> PagewalkResult:
        if entry is None:
            if pwndbg.aglib.memory.is_kernel(target):
                entry = pwndbg.aglib.regs.read_reg("TTBR1_EL1")
            else:
                entry = pwndbg.aglib.regs.read_reg("TTBR0_EL1")
        if entry is None:
            return PagewalkResult()
        entry |= 3  # marks the entry as a table
        return self._pagewalk(target, entry, virt)

    def pagescan(self, entry: int | None = None) -> list[Page]:
        # assumes entry should be from `kcurrent --set` and should be TTBR0_EL1 for a task
        if entry is None:
            entry = pwndbg.aglib.regs.read_reg("TTBR0_EL1")
        if entry is None:
            return []
        result = self._pagescan(entry | 3, is_kernel=False)
        if pwndbg.aglib.memory.is_kernel(pwndbg.aglib.regs.pc):
            entry = pwndbg.aglib.regs.read_reg("TTBR1_EL1")
            if entry is None:
                return []
            result += self._pagescan(entry | 3, is_kernel=True)
        return result

    def bitflags(self, level: PageTableLevel) -> BitFlags:
        if level.level == 1 or not level.entry or self.should_stop_pagewalk(level.entry):
            # block or page
            return BitFlags([("UNX", 54), ("PNX", 53), ("AP", (6, 7))])
        return BitFlags([("UNX", 60), ("PNX", 59), ("AP", (61, 62))])

    def should_stop_pagewalk(self, entry: int) -> bool:
        return (entry & 1) == 0 or (entry & 3) == 1



from __future__ import annotations

import argparse

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.hexdump
from pwndbg.color import message
from pwndbg.commands import CommandCategory
from pwndbg.lib.config import PARAM_ZUINTEGER

pwndbg.config.add_param("hexdump-width", 16, "line width of hexdump command")
pwndbg.config.add_param("hexdump-bytes", 64, "number of bytes printed by hexdump command")
pwndbg.config.add_param(
    "hexdump-group-width",
    -1,
    "number of bytes grouped in hexdump command",
    help_docstring="If -1, the architecture's pointer size is used.",
)
pwndbg.config.add_param(
    "hexdump-group-use-big-endian",
    False,
    "use big-endian within each group of bytes in hexdump command",
    help_docstring="When `on`, use big-endian within each group of bytes. Only applies to raw bytes, not the ASCII part. "
    "See also hexdump-highlight-group-lsb.",
)
pwndbg.config.add_param(
    "hexdump-limit-mb",
    10,
    "the maximum size in megabytes (MB) `hexdump` will read",
    help_docstring="""Set the maximum size in megabytes (MB) that the `hexdump` command will attempt to read at once.
    Prevents GDB crashes due to excessive memory allocation requests.
    Set to 0 for unlimited (use with caution).""",
    param_class=PARAM_ZUINTEGER,
)


def address_or_module_name(s) -> int:
    addr_or_str: int | str = pwndbg.commands.sloppy_gdb_parse(s)
    if isinstance(addr_or_str, str):
        module_name = addr_or_str
        pages = list(filter(lambda page: module_name in page.objfile, pwndbg.aglib.vmmap.get()))
        if pages:
            return pages[0].vaddr
        raise argparse.ArgumentTypeError(f"Could not find pages for module {module_name}")
    if isinstance(addr_or_str, int):
        return addr_or_str
    raise argparse.ArgumentTypeError("Unknown hexdump argument type.")


def format_c(data: bytes) -> str:
    toks = [f"{b:#02x}" for b in data]
    lines = []
    for i in range(0, len(toks), 16):
        elem = ", ".join(toks[i : i + 16])
        lines.append(f"    {elem},")
    body = "\n".join(lines)
    return f"static const unsigned char data[] = {{\n{body}\n}};\n"


def format_py(data: bytes) -> str:
    lines = []
    for i in range(0, len(data), 16):
        seg = "".join(f"\\x{b:02x}" for b in data[i : i + 16])
        lines.append(f'    b"{seg}"')
    return "data = (\n{}\n)\n".format("\n".join(lines))


parser = argparse.ArgumentParser(
    description="Hexdumps data at the specified address or module name."
)
parser.add_argument(
    "address",
    type=address_or_module_name,
    nargs="?",
    default="$sp",
    help="Address or module name to dump",
)
parser.add_argument(
    "count",
    type=int,
    nargs="?",
    default=pwndbg.config.hexdump_bytes,
    help="Number of bytes to dump",
)
parser.add_argument(
    "-C",
    "--code",
    type=str,
    nargs="?",
    const="py",
    choices=("py", "c"),
    help="Output as Python or C code data definition (default: py)",
)


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
@pwndbg.commands.OnlyWhenRunning
def hexdump(
    address: str | int, count: int = int(pwndbg.config.hexdump_bytes), code: str | None = None
) -> None:
    if count <= 0:
        print(f"count must be larger than 0 (is {count}).")
        return

    if hexdump.repeat:
        address = hexdump.last_address
    else:
        hexdump.offset = 0

    address = int(address)
    if address > pwndbg.aglib.arch.ptrmask:
        new_address = address & pwndbg.aglib.arch.ptrmask
        print(
            message.warn("0x%x is larger than the maximum address, truncating to 0x%x instead"),
            address,
            new_address,
        )
        address = new_address

    if not pwndbg.aglib.memory.peek(address):
        print("Could not read memory at specified address")
        return

    count = max(int(count), 0)

    # Get the configured limit in MB
    limit_mb = int(pwndbg.config.hexdump_limit_mb)

    # Check if the limit is enabled (non-zero) and if the request count exceeds it
    if limit_mb > 0:
        limit_bytes = limit_mb * 1024 * 1024
        if count > limit_bytes:
            # Raise an error with the informative message
            raise ValueError(
                f"Hexdump count ({count}) exceeds the current limit of {limit_mb} MB.\n"
                f"Use 'set hexdump-limit-mb <new_limit_in_mb>' to increase the limit (or set to 0 for unlimited)."
            )

    width = int(pwndbg.config.hexdump_width)

    group_width = int(pwndbg.config.hexdump_group_width)
    group_width = pwndbg.aglib.typeinfo.ptrsize if group_width == -1 else group_width

    # TODO: What if arch endian is big, and use_big_endian is false?
    flip_group_endianness = (
        bool(pwndbg.config.hexdump_group_use_big_endian) and pwndbg.aglib.arch.endian == "little"
    )

    # The user may have input the start and end range to dump instead of the
    # starting address and the number of bytes to dump. If the address is above
    # some minimum address, and the count is larger than that address, we assume
    # this is the case and correct it
    if 0x10000 < address < count:
        count -= address

    try:
        data = pwndbg.aglib.memory.read(address, count, partial=True)
        hexdump.last_address = address + count
    except pwndbg.dbg_mod.Error as e:
        print(e)
        return

    if code:
        source = format_py(data) if code == "py" else format_c(data)
        print(source)
        hexdump.offset += len(data)
    else:
        result = pwndbg.hexdump.hexdump(
            data,
            address=address,
            width=width,
            group_width=group_width,
            flip_group_endianness=flip_group_endianness,
            offset=hexdump.offset,
        )
        for line in result:
            print(line)

    hexdump.offset += len(data)


hexdump.last_address = 0
hexdump.offset = 0



"""
Overrides GDB's up and down commands to be prettier and to follow
the decompiler integration.
"""

from __future__ import annotations

import argparse

import gdb

import pwndbg
import pwndbg.commands
import pwndbg.commands.context
import pwndbg.commands.decompiler_integration
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Select and print stack frame that called this one.")
parser.add_argument(
    "n", nargs="?", default=1, type=int, help="The number of stack frames to go up."
)


@pwndbg.commands.Command(parser, category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
def up(n=1) -> None:
    """
    Select and print stack frame that called this one.
    """
    f = gdb.selected_frame()

    for i in range(int(n)):
        if f.older():
            f = f.older()
    f.select()

    # workaround for #632
    gdb.execute("frame", to_string=True)

    bt = pwndbg.commands.context.context_backtrace(with_banner=False)
    print("\n".join(bt))

    if pwndbg.commands.decompiler_integration.should_autojump:
        pwndbg.commands.decompiler_integration.auto_jump()


parser = argparse.ArgumentParser(description="Select and print stack frame called by this one.")
parser.add_argument(
    "n", nargs="?", default=1, type=int, help="The number of stack frames to go down."
)


# Since we are redefining a gdb command, we also redefine the original aliases.
# These aliases ("do", "dow") are necessary to ensure consistency in the help system
# and to pass the test_consistent_help test, which verifies that all commands and their
# aliases are documented correctly. See issue #2961 for more details.
@pwndbg.commands.Command(parser, category=CommandCategory.MISC, aliases=["do", "dow"])
@pwndbg.commands.OnlyWhenRunning
def down(n=1) -> None:
    """
    Select and print stack frame called by this one.
    """
    f = gdb.selected_frame()

    for i in range(int(n)):
        if f.newer():
            f = f.newer()
    f.select()

    # workaround for #632
    gdb.execute("frame", to_string=True)

    bt = pwndbg.commands.context.context_backtrace(with_banner=False)
    print("\n".join(bt))

    if pwndbg.commands.decompiler_integration.should_autojump:
        pwndbg.commands.decompiler_integration.auto_jump()



from __future__ import annotations

import gdb


class ControlTUIWindow:
    _tui_window: gdb.TuiWindow
    _button_text: str = "[←]  [→]"
    # Map from command to the span of the button in the _button_text.
    # The span is represented as (start, end) where start is the index
    # of the first character of the button and end is the index of the
    # last character of the button to react to on mouse click.
    _button_spans = {"contextprev": (0, 2), "contextnext": (5, 7)}

    def __init__(self, tui_window: gdb.TuiWindow) -> None:
        self._tui_window = tui_window
        self._tui_window.title = "history"

    def close(self) -> None:
        pass

    def render(self) -> None:
        self._tui_window.write(self._button_text, True)

    def hscroll(self, num: int) -> None:
        pass

    def vscroll(self, num: int) -> None:
        pass

    def click(self, x: int, y: int, button: int) -> None:
        # button specifies which mouse button was used, whose values can be 1 (left), 2 (middle), or 3 (right).
        if button != 1:
            return

        for command, (start, end) in self._button_spans.items():
            start_x = start % self._tui_window.width
            end_x = end % self._tui_window.width
            start_y = start // self._tui_window.width
            end_y = end // self._tui_window.width
            if start_x <= x <= end_x and start_y <= y <= end_y:
                gdb.execute(command, to_string=True)
                break


if hasattr(gdb, "register_window_type"):
    gdb.register_window_type("pwndbg_control", ControlTUIWindow)



from __future__ import annotations

import argparse

import pwnlib
import pwnlib.asm
import pwnlib.context

import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Assemble shellcode into bytes")

parser.add_argument(
    "-f", "--format", default="hex", choices=["hex", "string"], type=str, help="Output format"
)

parser.add_argument(
    "--arch",
    choices=pwnlib.context.context.architectures.keys(),
    type=str,
    help="Target architecture",
)

input_group = parser.add_mutually_exclusive_group(required=True)

input_group.add_argument(
    "shellcode", default=[], nargs="*", type=str, help="Assembler code to assemble"
)

input_group.add_argument("-i", "--infile", default=None, type=str, help="Specify input file")


@pwndbg.commands.Command(parser, command_name="asm", category=CommandCategory.MISC)
def asm(shellcode: list[str], format: str, arch: str | None, infile: str) -> None:
    if infile:
        print(message.warn("Going to read from file: " + infile))
        with open(infile) as file:
            shellcode = [file.read()]

    if arch is None:
        arch = pwnlib.context.context.arch

    bits_for_arch = pwnlib.context.context.architectures.get(arch, {}).get("bits")
    assembly = pwnlib.asm.asm(" ".join(shellcode), arch=arch, bits=bits_for_arch)

    if format == "hex":
        print(assembly.hex())
    else:
        print(assembly)



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.aglib.asm
import pwndbg.aglib.shellcode
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.commands import CommandCategory

# Taken from linux/arch/x86/include/asm/msr-index.h
# Arbitrary list, feel free to add more.
X86_MSRS = {
    "MSR_EFER": 0xC0000080,
    "MSR_STAR": 0xC0000081,
    "MSR_LSTAR": 0xC0000082,
    "MSR_CSTAR": 0xC0000083,
    "MSR_SYSCALL_MASK": 0xC0000084,
    "MSR_FS_BASE": 0xC0000100,
    "MSR_GS_BASE": 0xC0000101,
    "MSR_KERNEL_GS_BASE": 0xC0000102,
    "MSR_TSC_AUX": 0xC0000103,
}


COMMON_MSRS = {"i386": X86_MSRS, "x86-64": X86_MSRS}


def parse_msr(msr: str, arch: str) -> int | None:
    # first try to parse MSR name, then as int/hex
    if (val := COMMON_MSRS.get(arch, {}).get(msr.upper())) is not None:
        return val

    try:
        return int(msr, 0)
    except ValueError:
        print(f"unknown MSR {msr} on {arch}")
        return None


def parse_range(msr_range: str, arch: str) -> tuple[int, int] | None:
    bounds = msr_range.split("-")
    if len(bounds) != 2:
        return None

    start = parse_msr(bounds[0], arch)
    end = parse_msr(bounds[1], arch)

    if start is None or end is None:
        return None

    if start > end:
        return None

    return (start, end)


def x86_msr_read(msr: int) -> None:
    async def ctrl(ec: pwndbg.dbg_mod.ExecutionController):
        sc = pwndbg.aglib.asm.asm(f"mov ecx, {msr}; rdmsr")
        async with pwndbg.aglib.shellcode.exec_shellcode(ec, sc):
            edx = int(pwndbg.aglib.regs.read_reg("edx")) << 32
            eax = int(pwndbg.aglib.regs.read_reg("eax"))
            ret = edx + eax
            print(f"{hex(msr)}:\t{hex(ret)}")

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(ctrl)


def x86_msr_write(msr: int, write_value: int) -> None:
    async def ctrl(ec: pwndbg.dbg_mod.ExecutionController):
        eax = write_value & 0xFFFFFFFF
        edx = write_value >> 32
        sc = pwndbg.aglib.asm.asm(f"mov ecx, {msr}; mov eax, {eax}; mov edx, {edx}; wrmsr")
        async with pwndbg.aglib.shellcode.exec_shellcode(ec, sc):
            return

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(ctrl)


def msr_read(msr: int) -> None:
    arch = pwndbg.aglib.arch.name

    if arch == "i386" or arch == "x86-64":
        x86_msr_read(msr)
    else:
        print(f"{arch} not supported")


def msr_write(msr: int, write_value: int) -> None:
    arch = pwndbg.aglib.arch.name

    if arch == "i386" or arch == "x86-64":
        x86_msr_write(msr, write_value)
    else:
        print(f"{arch} not supported")


def msr_list(arch: str):
    msrs = COMMON_MSRS[arch]
    longest_msr_name = max(map(len, msrs))
    for msr, value in msrs.items():
        print(f"{msr.ljust(longest_msr_name)} = {hex(value)}")


parser = argparse.ArgumentParser(
    description="""
Read or write to Model Specific Register (MSR)
""",
)
parser.add_argument("msr", help="MSR value or name", type=str, nargs="?", default=None)
parser.add_argument(
    "-w",
    "--write",
    metavar="write_value",
    help="value to write in MSR",
    type=int,
    nargs="?",
    default=None,
)
parser.add_argument(
    "-l",
    "--list",
    dest="list_msr",
    action="store_true",
    help="list common MSRs for the current arch",
    default=False,
)
parser.add_argument(
    "-r",
    "--range",
    dest="msr_range",
    help="dash separated range of MSRs to read (eg. --range=1-10 where 10 is included)",
    type=str,
    nargs="?",
    default=None,
)


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
def msr(
    msr: str | None = None,
    write: int | None = None,
    list_msr=False,
    msr_range: str | None = None,
) -> None:
    arch = pwndbg.aglib.arch.name

    if msr:
        msr_val = parse_msr(msr, arch)
        if msr_val is None:
            print(f"Error: failed to parse {msr}")
            return

        if write:
            msr_write(msr_val, write)
            return

        msr_read(msr_val)

    elif msr_range:
        parsed_range = parse_range(msr_range, arch)
        if not parsed_range:
            print("Error: invalid range")
            return

        for m in range(parsed_range[0], parsed_range[1] + 1):
            msr_read(m)

    elif list_msr:
        msr_list(arch)

    else:
        parser.print_usage()



from __future__ import annotations

import contextlib
from collections.abc import Iterator

import pwnlib.elf.elf
import pwnlib.term.text
from pwnlib.elf import ELF

pwnlib.elf.elf.log._logger.disabled = True

import pwndbg.color.message as message


@contextlib.contextmanager
def monkeypatch_pwnlib_term_text() -> Iterator[None]:
    # Note: It's kinda hacky to monkeypatch pwnlib.term.text like this, but I didn't find a better way to do it.
    # The patch is for here:
    # https://github.com/Gallopsled/pwntools/blob/f046fdd93e154bd892332f38cfbb518de130f1f2/pwnlib/elf/elf.py#L1999-L2001
    # This might break in the future, so need to update this patch when the implementation of pwnlib changes.
    pwnlib.term.text.red = message.error
    pwnlib.term.text.green = message.success
    pwnlib.term.text.yellow = message.warn
    yield
    del pwnlib.term.text.red
    del pwnlib.term.text.green
    del pwnlib.term.text.yellow


def get_raw_out(local_path: str) -> str:
    elf = ELF(local_path)
    # 10 is the magic number used in elf.checksec() to align the output.
    # https://github.com/Gallopsled/pwntools/blob/f046fdd93e154bd892332f38cfbb518de130f1f2/pwnlib/elf/elf.py#L2012
    # We might need to update this number if the implementation of elf.checksec() changes in the future.
    output = "File:".ljust(10) + elf.path + "\n"
    output += "Arch:".ljust(10) + elf.arch + "\n"
    with monkeypatch_pwnlib_term_text():
        output += elf.checksec()
    return output


def relro_status(local_path: str) -> str:
    return {"Full": "Full RELRO", "Partial": "Partial RELRO", None: "No RELRO"}[
        ELF(local_path).relro
    ]


def pie_status(local_path: str) -> str:
    return "PIE enabled" if ELF(local_path).pie else "No PIE"



"""
Functionality for disassmebling code at an address, or at an
address +/- a few instructions.
"""

from __future__ import annotations

import collections
import re
from collections.abc import Callable
from dataclasses import dataclass

from capstone6pwndbg import *  # noqa: F403

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.disasm.aarch64
import pwndbg.aglib.disasm.arch
import pwndbg.aglib.disasm.arm
import pwndbg.aglib.disasm.loongarch64
import pwndbg.aglib.disasm.mips
import pwndbg.aglib.disasm.ppc
import pwndbg.aglib.disasm.riscv
import pwndbg.aglib.disasm.sparc
import pwndbg.aglib.disasm.x86
import pwndbg.aglib.memory
import pwndbg.emu.emulator
import pwndbg.lib.cache
import pwndbg.lib.config
from pwndbg.aglib.disasm.arch import DEBUG_ENHANCEMENT
from pwndbg.aglib.disasm.arch import DisassemblyAssistant
from pwndbg.aglib.disasm.instruction import ManualPwndbgInstruction
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.aglib.disasm.instruction import PwndbgInstructionImpl
from pwndbg.aglib.disasm.instruction import SplitType
from pwndbg.color import message
from pwndbg.dbg_mod import EventType
from pwndbg.lib.arch import PWNDBG_SUPPORTED_ARCHITECTURES_TYPE

CapstoneSyntax = {"intel": CS_OPT_SYNTAX_INTEL, "att": CS_OPT_SYNTAX_ATT}

force_register_alias = pwndbg.config.add_param(
    "disasm-reg-alias",
    False,
    "force the disassembly to use register aliases (e.g. aarch64 x29 -> fp)",
    param_class=pwndbg.lib.config.PARAM_BOOLEAN,
    help_docstring="""\
The register aliasing is done by capstone, see:
https://github.com/capstone-engine/capstone/blob/next/docs/cs_v6_release_guide.md#:~:text=None.-,Register%20alias,-Register%20alias%20

Enabling this may make disassembly slower.
""",
)

# Caching strategy:
# To ensure we don't have stale register/memory information in our cached PwndbgInstruction,
# we clear the cache whenever we DON'T do a `stepi`, `nexti`, `step`, or `next` command.
# Although `stepi` and `nexti` always go to the next machine instruction in memory, `step` and `next`
# can skip over multiple when GDB has debugging symbols and sourcecode
# In order to determine that we did a `stepi`, `nexti`, `step`, or `next`, whenever the process stops,
# we check if the current program counter is at the address of one of the instructions that we
# emulated to the last time the process stopped. This allows use to skips a handful of instruction, but still retain the cache
# Any larger changes of the program counter will cause the cache to reset.

next_addresses_cache: set[int] = set()

# The disassembly system isn't able to remember that an instruction is a delay slot instruction when it is disassembled in isolation
# from the branch is belongs to.
# This cache is used to handle this. Each address points to the branch that created the delay slot.
delay_slot_cache: collections.defaultdict[int, PwndbgInstruction | None] = collections.defaultdict(
    lambda: None
)


# Register GDB event listeners for all stop events
@pwndbg.dbg.event_handler(EventType.STOP)
def enhance_cache_listener() -> None:
    if pwndbg.aglib.regs.pc not in next_addresses_cache:
        # Clear the enhanced instruction cache to ensure we don't use stale values
        computed_instruction_cache.clear()
        instruction_sequence_linked_list_map.clear()


@pwndbg.dbg.event_handler(EventType.MEMORY_CHANGED)
@pwndbg.dbg.event_handler(EventType.REGISTER_CHANGED)
def clear_on_reg_mem_change() -> None:
    # We clear all the future computed instructions because when we manually change a register or memory, it's often a location
    # used by the instructions at or just after the current PC, and our previously emulated future instructions might be inaccurate
    computed_instruction_cache.pop(pwndbg.aglib.regs.pc, None)
    instruction_sequence_linked_list_map.pop(pwndbg.aglib.regs.pc, None)

    for addr in next_addresses_cache:
        computed_instruction_cache.pop(addr, None)
        instruction_sequence_linked_list_map.pop(addr, None)

    next_addresses_cache.clear()


# Dict of Address -> previous instruction sequentially in memory
# Some architectures don't have fixed-sized instructions, so this is used
# to disassemble backwards linearly in memory for those cases
linear_backward_cache: collections.defaultdict[int, int] = collections.defaultdict(lambda: None)


# In order to track the sequence of instructions at runtime, we maintain a linked list, where each
# entry points to the previous instruction that was executed.
# This is populated speculatively using emulation.
@dataclass
class InstructionSequenceNode:
    """This is used to form a linked list that tracks the order of instructions execution at runtime"""

    previous: InstructionSequenceNode | None
    instruction: PwndbgInstruction


@dataclass
class InstructionSequenceSavePointer:
    """
    This allows preserving context across calls when disassembling backwards.
    It stores the node that we just processed internally.

    This is like the saveptr parameter of strtok_r.
    """

    node: InstructionSequenceNode | None


# Map addresses to their entry in the linked list.
# While the emulation may encounter this address multiple times, this map only contains a mapping for the first
# time the instruction is executed.
instruction_sequence_linked_list_map: collections.defaultdict[
    int, InstructionSequenceNode | None
] = collections.defaultdict(lambda: None)


# In case the linked list method fails (we cannot be 100% certain of instruction order),
# it is still nice to be able to display instructions behind the instruction pointer.
# This tracks the order of instructions based on the last time we disassembled them.
# Map of address to previous address
fallback_backward_cache: collections.defaultdict[int, int | None] = collections.defaultdict(
    lambda: None
)

# This allows use to retain the annotation strings from previous instructions
computed_instruction_cache: collections.defaultdict[int, PwndbgInstruction | None] = (
    collections.defaultdict(lambda: None)
)

# Maps an address to integer 0/1, indicating the Thumb mode bit for the given address.
# Value is None if Thumb bit is irrelevent or unknown.
emulated_arm_mode_cache: collections.defaultdict[int, int | None] = collections.defaultdict(
    lambda: None
)


def get_instruction_sequence_node(
    address: int, saveptr: InstructionSequenceSavePointer
) -> InstructionSequenceNode | None:
    """Return the node of the linked list at the given address, if it exists"""
    if saveptr.node is not None:
        return saveptr.node

    if (val := instruction_sequence_linked_list_map.get(address)) is not None:
        return val

    return None


def get_previous_instruction(
    address: int, use_cache: bool, linear: bool, saveptr: InstructionSequenceSavePointer
) -> PwndbgInstruction | None:
    """
    Retrieve the instruction prior to the instruction at `address`.
    """
    if linear:
        prev_address = linear_backward_cache[address]
        result = (
            one(prev_address, from_cache=use_cache, put_backward_cache=False, linear=linear)
            if prev_address
            else None
        )
        if result is None and pwndbg.aglib.arch.constant_instruction_size:
            return one(
                address - pwndbg.aglib.arch.max_instruction_size,
                from_cache=use_cache,
                put_backward_cache=False,
                linear=linear,
            )
        return result

    sequence_node = get_instruction_sequence_node(address, saveptr)

    if sequence_node is not None:
        prev_node = sequence_node.previous
        saveptr.node = prev_node
        if prev_node is not None:
            return prev_node.instruction

    prev_address = fallback_backward_cache[address]
    return (
        one(prev_address, from_cache=use_cache, put_backward_cache=False) if prev_address else None
    )


@pwndbg.lib.cache.cache_until("objfile")
def get_disassembler(cs_info: tuple[int, int]) -> Cs:
    arch, mode = cs_info

    mode |= pwndbg.aglib.arch.get_capstone_endianness()

    cs = Cs(arch, mode)

    flavor = pwndbg.dbg.x86_disassembly_flavor()
    try:
        cs.syntax = CapstoneSyntax[flavor]
        if force_register_alias:
            cs.syntax |= CS_OPT_SYNTAX_CS_REG_ALIAS
    except CsError:
        pass
    cs.detail = True
    return cs


def one(
    address: int | None = None,
    emu: pwndbg.emu.emulator.Emulator = None,
    enhance: bool = True,
    assistant: DisassemblyAssistant | None = None,
    from_cache: bool = False,
    put_cache: bool = False,
    put_backward_cache: bool = True,
    linear: bool = False,
) -> PwndbgInstruction | None:
    """
    Return None on failure to fetch an instruction
    """

    if address is None:
        address = pwndbg.aglib.regs.pc

    if not pwndbg.aglib.memory.peek(address):
        return None

    if from_cache:
        cached = computed_instruction_cache[address]
        if cached is not None:
            return cached

    if (
        insn := get_one_instruction(address, emu, enhance=enhance, assistant=assistant)
    ) is not None:
        if put_cache:
            computed_instruction_cache[address] = insn

        if put_backward_cache:
            linear_backward_cache[insn.address + insn.size] = insn.address

            if not linear:
                fallback_backward_cache[insn.next] = insn.address
        return insn

    return None


# Get one instruction without enhancement
def one_raw(address=None) -> PwndbgInstruction | None:
    if address is None:
        address = pwndbg.aglib.regs.pc

    if not pwndbg.aglib.memory.peek(address):
        return None

    return get_one_instruction(address, enhance=False)


def get(
    address: int,
    instructions: int = 1,
    emu: pwndbg.emu.emulator.Emulator | None = None,
    enhance: bool = True,
    assistant: DisassemblyAssistant | None = None,
    padding: int = 6,
) -> list[PwndbgInstruction]:
    address = int(address)

    # Dont disassemble if there's no memory
    if not pwndbg.aglib.memory.peek(address):
        return []

    retval: list[PwndbgInstruction] = []
    for _ in range(instructions):
        i = get_one_instruction(
            address,
            emu,
            enhance=enhance,
            assistant=assistant,
            padding=padding,
        )
        if i is None:
            break
        address = i.next
        retval.append(i)

    return retval


def get_one_instruction(
    address: int,
    emu: pwndbg.emu.emulator.Emulator | None = None,
    enhance: bool = True,
    assistant: DisassemblyAssistant | None = None,
    padding: int = 6,
) -> PwndbgInstruction | None:
    """
    If passed an emulator, this will pass it to the DisassemblyAssistant which will
    single_step the emulator to determine the operand values before and after the instruction executes.
    """
    cs_info = pwndbg.aglib.arch.get_capstone_constants(address)
    if cs_info is None:
        instr = ManualPwndbgInstruction(address, padding)
        if enhance:
            pwndbg.aglib.disasm.arch.basic_enhance(instr)
        return instr

    md = get_disassembler(cs_info)
    data = pwndbg.aglib.memory.read(address, pwndbg.aglib.arch.max_instruction_size, partial=True)
    for ins in md.disasm(bytes(data), address, 1):
        pwn_ins: PwndbgInstruction = PwndbgInstructionImpl(ins, padding)

        if enhance:
            if assistant is None:
                assistant = (
                    pwndbg.aglib.disasm.disassembly.get_disassembly_assistant_for_current_arch()
                )
            assistant.enhance(pwn_ins, emu)

        return pwn_ins

    # This is reached if disassembly fails (unknown sequence of bytes)
    return None


def can_run_first_emulate() -> bool:
    """
    Disable the emulate config variable if we don't have enough memory to use it
    See https://github.com/pwndbg/pwndbg/issues/1534
    And https://github.com/unicorn-engine/unicorn/pull/1743
    """
    global first_time_emulate
    if not first_time_emulate:
        return True
    first_time_emulate = False

    try:
        from mmap import mmap

        mm = mmap(-1, 1024 * 1024 * 1024)
        mm.close()
    except OSError:
        print(
            message.error(
                "Disabling the emulation via Unicorn Engine that is used for computing branches"
                " as there isn't enough memory (1GB) to use it (since mmap(1G, RWX) failed). See also:\n"
                "* https://github.com/pwndbg/pwndbg/issues/1534\n"
                "* https://github.com/unicorn-engine/unicorn/pull/1743\n"
                "Either free your memory or explicitly set `set emulate off` in your Pwndbg config"
            )
        )
        pwndbg.config.emulate.value = "off"
        return False

    return True


first_time_emulate = True


def no_emulate_one():
    result = near(pwndbg.aglib.regs.pc, emulate=False, show_prev_insns=False)
    if result:
        return result[0][0]
    return None


def emulate_one():
    result = near(pwndbg.aglib.regs.pc, emulate=True, show_prev_insns=False)
    if result:
        return result[0][0]
    return None


def one_with_config():
    """
    Returns a single Pwndbg Instruction at the current PC.

    Emulation determined by the `pwndbg.config.emulate` setting.
    """
    result = near(
        pwndbg.aglib.regs.pc,
        emulate=bool(not pwndbg.config.emulate == "off"),
        show_prev_insns=False,
    )
    if result:
        return result[0][0]
    return None


def set_visual_split(
    set_ins: PwndbgInstruction, check_ins: PwndbgInstruction, linear: bool
) -> None:
    """
    Internal helper function to set the .split property for display purposes.

    This should only be called when the callee knows that a split should be created.

    set_ins is the instruction that we are modifying

    checks_ins is the one used to check what type of split is necessary.
    The same as set_ins unless it's a delay slot.
    """
    if not linear and (
        check_ins.next != check_ins.address + check_ins.size
        or check_ins.force_unconditional_jump_target
    ):
        set_ins.split = SplitType.BRANCH_TAKEN
    else:
        set_ins.split = SplitType.BRANCH_NOT_TAKEN


# Return (list of PwndbgInstructions, index in list where instruction.address = passed in address)
def near(
    address: int,
    forward_count: int = 1,
    backward_count: int = 0,
    total_count: int | None = None,
    end_address: int | None = None,
    emulate=False,
    show_prev_insns=True,
    use_cache=False,
    linear=False,
) -> tuple[list[PwndbgInstruction], int]:
    """
    Disassembles instructions near given `address`. Passing `emulate` makes use of
    unicorn engine to emulate instructions to predict branches that will be taken.
    `show_prev_insns` makes this show previously cached instructions

    This allows us to maintain a context of surrounding instructions while
    single-stepping instructions.

    Args:
        forward_count: number of instructions forward from this instruction
        backward_count: maximum number of previously executed instructions
        total_count:
            if set, returns a list with this many instructions in total.
            The number of backward instructions is limited by `backward_count`.
            If this is set, `forward_count` is ignored.
        end_address:
            determines the maximum address (non-inclusive) that can be disassembled.
    """

    pc = pwndbg.aglib.regs.pc

    # Some architecture aren't emulated yet
    if not pwndbg.emu or pwndbg.aglib.arch.name not in pwndbg.emu.emulator.arch_to_UC:
        emulate = False

    emu: pwndbg.emu.emulator.Emulator = None

    # Emulate if program pc is at the current instruction - can't emulate at arbitrary places, because we need current
    # processor state to instantiate the emulator.
    if address == pc and emulate and (not first_time_emulate or can_run_first_emulate()):
        try:
            emu = pwndbg.emu.emulator.Emulator()
        except pwndbg.dbg_mod.Error as e:
            match = re.search(r"Memory at address (\w+) unavailable\.", str(e))
            if match:
                return ([], -1)
            raise

    # By using the same assistant for all the instructions disassembled in this pass, we can track and share information across the instructions
    assistant = pwndbg.aglib.disasm.disassembly.get_disassembly_assistant_for_current_arch()

    # Copy register values to the enhancer for use in manual register tracking
    if assistant.supports_manual_emulation and address == pc:
        for reg in pwndbg.aglib.regs.current.common:
            if (reg_value := pwndbg.aglib.regs.read_reg(reg)) is not None:
                assistant.manual_register_values.write_register(reg, reg_value)

    # Start at the current instruction using emulation if available.
    current = one(address, emu, put_cache=True, assistant=assistant, linear=linear)

    if DEBUG_ENHANCEMENT:
        if emu and not emu.last_step_succeeded:
            print("Emulator failed at first step")

    if current is None:
        return ([], -1)

    # A linked list that contains the order of instructions that emulation
    # determines will run upon uses of the "nexti" command.
    instruction_sequence_head = instruction_sequence_linked_list_map.get(address)

    if instruction_sequence_head is None:
        instruction_sequence_head = InstructionSequenceNode(None, current)
        instruction_sequence_linked_list_map[address] = instruction_sequence_head
    else:
        # We re-disassembled the instruction and enhanced it, so save the new value
        instruction_sequence_head.instruction = current

    insns: list[PwndbgInstruction] = []

    # Get previously executed instructions from the cache.
    if DEBUG_ENHANCEMENT:
        print(f"CACHE START -------------------, {current.address}")

    if show_prev_insns:
        saveptr = InstructionSequenceSavePointer(None)

        insn = get_previous_instruction(
            current.address, use_cache=use_cache, linear=linear, saveptr=saveptr
        )
        while insn is not None and len(insns) < backward_count:
            if DEBUG_ENHANCEMENT:
                print(f"Got instruction from cache, addr={insn.address:#x}")
            if insn.jump_like and insn.split == SplitType.NO_SPLIT and not insn.causes_branch_delay:
                insn.split = SplitType.BRANCH_NOT_TAKEN
            insns.append(insn)

            insn = get_previous_instruction(
                insn.address, use_cache=use_cache, linear=linear, saveptr=saveptr
            )
        insns.reverse()

    if total_count is not None:
        target_instruction_count = total_count
    else:
        target_instruction_count = len(insns) + forward_count

    index_of_current_instruction = len(insns)
    insns.append(current)

    if DEBUG_ENHANCEMENT:
        print("END CACHE -------------------")

    # At this point, we've already added everything *BEFORE* the requested address,
    # and the instruction at 'address'.
    # Now, continue forwards.

    # A set of all the addresses after the PC that we have disassembled in this pass
    new_addresses_seen: set[int] = set()

    next_addresses_cache.clear()
    next_addresses_cache.add(current.target)

    insn = current

    last_emulated_thumb_bit_value: int | None = None

    while insn and len(insns) < target_instruction_count:
        target = insn.next if not linear else insn.address + insn.size

        if end_address is not None and target >= end_address:
            break

        # Emulation may have failed or been disabled in the last call to one()
        if emu:
            if not emu.last_step_succeeded or not emu.valid:
                emu = None
            else:
                # Upon execution the previous instruction, the Thumb mode bit may have changed.
                # This means we know whether the next instruction executed will be Thumb or not.
                # This returns None in the case the Thumb bit is not relevent.
                last_emulated_thumb_bit_value = emulated_arm_mode_cache[emu.pc()] = (
                    emu.read_thumb_bit()
                )

        if not emu and last_emulated_thumb_bit_value is not None:
            # The emulator may have been disabled, but while it was live we transitioned into Thumb mode.
            # We propagate the Thumb mode through the remaining instructions we disassemble.
            emulated_arm_mode_cache[target] = last_emulated_thumb_bit_value

        # Handle visual splits in the disasm view
        # We create splits in 3 conditions:
        # 1. We know the instruction is "jump_like" - it mutates the PC. We don't necessarily know the target, but know it can have one.
        # 2. The instruction has an explicitly resolved target which is not the next instruction in memory
        # 3. The instruction repeats (like x86 `REP`)
        split_insn = insn
        if insn.jump_like or insn.has_jump_target or insn.next == insn.address:
            # This branch handles delay slots. Delay slots have an interesting quirk in debuggers:
            # sometimes the debugger can pause in the delay slot, and sometimes the debugger will
            # automatically step over it.
            if insn.causes_branch_delay:
                # Delay slots are instructions after branches that always execute.
                # They offer a couple challenges because debuggers the Unicorn often cannot be paused in a delay slot instruction.
                # Single stepping on a branch will cause the debugger to execute the delay slot instruction and take the branch action.
                # This means the emulator's program counter will take on the value that the branch action dictates, and we would normally continue disassembling there.
                # Therefore, we must disassemble the delay slot instructions here as the normal codeflow will not reach them.

                delay_slot_address = insn.address + insn.size
                split_insn = one(delay_slot_address, None, put_cache=True, linear=linear)

                # There might not be a valid instruction at the branch delay slot
                if split_insn is None:
                    break

                insns.append(split_insn)

                ### Start manually handling caching related to delay slots
                next_addresses_cache.add(split_insn.address)

                delay_slot_cache[split_insn.address] = insn

                fallback_backward_cache[insn.next] = split_insn.address
                fallback_backward_cache[split_insn.address + split_insn.size] = split_insn.address
                fallback_backward_cache[split_insn.address] = insn.address

                instruction_sequence_head = InstructionSequenceNode(
                    instruction_sequence_head, split_insn
                )

                if delay_slot_address not in new_addresses_seen:
                    new_addresses_seen.add(delay_slot_address)
                    instruction_sequence_linked_list_map[delay_slot_address] = (
                        instruction_sequence_head
                    )
                ### Done handling caching stuff

                # Because the emulator failed, we manually set the address of the next instruction.
                # This is the address that typing "nexti" in GDB will take us to
                target = split_insn.address + split_insn.size

                if not insn.call_like and (
                    insn.is_unconditional_jump or insn.is_conditional_jump_taken
                ):
                    target = insn.target

            set_visual_split(split_insn, insn, linear)

        # Handle edge case where debugger is paused on the delay slot instruction
        # Force the disassembly flow to follow the direction of the branch
        if (cached_ins := delay_slot_cache[insn.address]) is not None:
            if not cached_ins.call_like and (
                cached_ins.is_unconditional_jump or cached_ins.is_conditional_jump_taken
            ):
                target = insn.next = cached_ins.next

            set_visual_split(insn, cached_ins, linear)

        next_addresses_cache.add(target)

        # The emulator is stepped within this call
        insn = one(target, emu, put_cache=True, assistant=assistant, linear=linear)

        if insn:
            # Add the instruction to the front of the linked list tracking the dynamic instruction sequence.
            instruction_sequence_head = InstructionSequenceNode(instruction_sequence_head, insn)

            # We want to add for the first time an instruction is encountered
            # in the current disassembly flow.
            if target not in new_addresses_seen:
                new_addresses_seen.add(target)
                instruction_sequence_linked_list_map[target] = instruction_sequence_head

            insns.append(insn)

    # Remove repeated instructions at the end of disassembly.
    # Always ensure we display the current and *next* instruction,
    # but any repeats after that are removed.
    #
    # This helps with infinite loops and RET sleds.

    while insns and len(insns) > 2 and insns[-3].address == insns[-2].address == insns[-1].address:
        del insns[-1]

    return (insns, index_of_current_instruction)


ALL_DISASSEMBLY_ASSISTANTS: dict[
    PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, Callable[[], DisassemblyAssistant]
] = {
    "aarch64": lambda: pwndbg.aglib.disasm.aarch64.AArch64DisassemblyAssistant("aarch64"),
    "i386": lambda: pwndbg.aglib.disasm.x86.X86DisassemblyAssistant("i386"),
    "x86-64": lambda: pwndbg.aglib.disasm.x86.X86DisassemblyAssistant("x86-64"),
    "arm": lambda: pwndbg.aglib.disasm.arm.ArmDisassemblyAssistant("arm", "cpsr"),
    "armcm": lambda: pwndbg.aglib.disasm.arm.ArmDisassemblyAssistant("armcm", "xpsr"),
    "mips": lambda: pwndbg.aglib.disasm.mips.MipsDisassemblyAssistant("mips"),
    "rv32": lambda: pwndbg.aglib.disasm.riscv.RISCVDisassemblyAssistant("rv32"),
    "rv64": lambda: pwndbg.aglib.disasm.riscv.RISCVDisassemblyAssistant("rv64"),
    "loongarch64": lambda: pwndbg.aglib.disasm.loongarch64.Loong64DisassemblyAssistant(
        "loongarch64"
    ),
    "powerpc": lambda: pwndbg.aglib.disasm.ppc.PowerPCDisassemblyAssistant("powerpc"),
    "sparc": lambda: pwndbg.aglib.disasm.sparc.SparcDisassemblyAssistant("sparc"),
}


def get_disassembly_assistant_for_current_arch() -> DisassemblyAssistant:
    # If a specific subclass has not been created for the given arc, return the generic assistant
    return ALL_DISASSEMBLY_ASSISTANTS.get(
        pwndbg.aglib.arch.name, lambda: DisassemblyAssistant(None)
    )()


def arch_has_disassembly_assistant(arch: PWNDBG_SUPPORTED_ARCHITECTURES_TYPE | None = None) -> bool:
    if arch is None:
        arch = pwndbg.aglib.arch.name

    return arch in ALL_DISASSEMBLY_ASSISTANTS



"""
Text colorizing and theme configuration logic.
"""

from __future__ import annotations

import os
import re
from collections.abc import Callable
from typing import NamedTuple

import pwndbg
from pwndbg.lib.config import Parameter

from . import theme

NORMAL = "\x1b[0m"
BLACK = "\x1b[30m"
RED = "\x1b[31m"
GREEN = "\x1b[32m"
YELLOW = "\x1b[33m"
BLUE = "\x1b[34m"
PURPLE = "\x1b[35m"
CYAN = "\x1b[36m"
LIGHT_GREY = LIGHT_GRAY = "\x1b[37m"
FOREGROUND = "\x1b[39m"
GREY = GRAY = "\x1b[90m"
LIGHT_RED = "\x1b[91m"
LIGHT_GREEN = "\x1b[92m"
LIGHT_YELLOW = "\x1b[93m"
LIGHT_BLUE = "\x1b[94m"
LIGHT_PURPLE = "\x1b[95m"
LIGHT_CYAN = "\x1b[96m"
WHITE = "\x1b[97m"
BOLD = "\x1b[1m"
UNDERLINE = "\x1b[4m"


# We assign `none` instead of creating a function since it is faster this way
# While this is a microptimization, the `none` may be called thousands of times with
# a single context or a `hexdump $rsp 5000` call
# A simple benchmark below:
#   In [1]: def f(x): return str(x)
#   In [2]: %timeit f('')
#   117 ns ± 0.642 ns per loop (mean ± std. dev. of 7 runs, 10000000 loops each)
#   In [3]: %timeit str('')
#   72 ns ± 0.222 ns per loop (mean ± std. dev. of 7 runs, 10000000 loops each)
none = str


def normal(x: str) -> str:
    return colorize(x, NORMAL)


def black(x: str) -> str:
    return colorize(x, BLACK)


def red(x: str) -> str:
    return colorize(x, RED)


def green(x: str) -> str:
    return colorize(x, GREEN)


def yellow(x: str) -> str:
    return colorize(x, YELLOW)


def blue(x: str) -> str:
    return colorize(x, BLUE)


def purple(x: str) -> str:
    return colorize(x, PURPLE)


def cyan(x: str) -> str:
    return colorize(x, CYAN)


def light_gray(x: str) -> str:
    return colorize(x, LIGHT_GRAY)


def foreground(x: str) -> str:
    return colorize(x, FOREGROUND)


def gray(x: str) -> str:
    return colorize(x, GRAY)


def light_red(x: str) -> str:
    return colorize(x, LIGHT_RED)


def light_green(x: str) -> str:
    return colorize(x, LIGHT_GREEN)


def light_yellow(x: str) -> str:
    return colorize(x, LIGHT_YELLOW)


def light_blue(x: str) -> str:
    return colorize(x, LIGHT_BLUE)


def light_purple(x: str) -> str:
    return colorize(x, LIGHT_PURPLE)


def light_cyan(x: str) -> str:
    return colorize(x, LIGHT_CYAN)


def white(x: str) -> str:
    return colorize(x, WHITE)


def bold(x: str) -> str:
    return colorize(x, BOLD)


def underline(x: str) -> str:
    return colorize(x, UNDERLINE)


def colorize(x: str, color: str) -> str:
    return color + terminateWith(str(x), color) + NORMAL


def nocolor(x: str, color: str) -> str:
    return x


# Taken from https://stackoverflow.com/a/14693789
ansi_escape_8bit = re.compile(
    r"(?:\x1B[@-Z\\-_]|[\x80-\x9A\x9C-\x9F]|(?:\x1B\[|\x9B)[0-?]*[ -/]*[@-~])"
)


def unstylize(x: str) -> str:
    return ansi_escape_8bit.sub("", x)


disable_colors = theme.add_param(
    "disable-colors",
    bool(os.environ.get("NO_COLOR")),
    "whether to color the output or not",
)


@pwndbg.config.trigger(disable_colors)
def _disable_colors_trigger():
    if disable_colors:
        if not hasattr(colorize, "original_code"):
            colorize.original_code = colorize.__code__
        colorize.__code__ = nocolor.__code__
    else:
        if hasattr(colorize, "original_code"):
            colorize.__code__ = colorize.original_code


def generateColorFunctionInner(
    old: Callable[[object], str], new: Callable[[str], str]
) -> Callable[[object], str]:
    def wrapper(text: object) -> str:
        return new(old(text))

    return wrapper


class ColorParamSpec(NamedTuple):
    name: str
    default: str
    doc: str


class ColorConfig:
    def __init__(self, namespace: str, params: list[ColorParamSpec]) -> None:
        self._namespace = namespace
        self._params: dict[str, theme.ColorParameter] = {}
        for param in params:
            self._params[param.name] = theme.add_color_param(
                f"{self._namespace}-{param.name}-color", param.default, param.doc
            )

    def __getattr__(self, attr: str) -> Callable[[str], str]:
        param_name = attr.replace("_", "-")
        if param_name in self._params:
            return self._params[param_name].color_function

        raise AttributeError(f"ColorConfig object for {self._namespace} has no attribute '{attr}'")


def generateColorFunction(
    config: str | Parameter, _globals: dict[str, Callable[[str], str]] = globals()
) -> Callable[[object], str]:
    # the `config` here may be a config Parameter object
    # and if we run with disable_colors or if the config value
    # is empty, we need to ensure we cast it to string
    # so it can be properly formatted e.g. with:
    # "{config_param:5}".format(config_param=some_config_parameter)
    function = str

    if disable_colors:
        return function

    for color in config.split(","):
        func_name = color.lower().replace("-", "_")
        function = generateColorFunctionInner(function, _globals[func_name])
    return function


def strip(x: str) -> str:
    return re.sub("\x1b\\[[\\d;]+m", "", x)


def terminateWith(x: str, color: str) -> str:
    return x.replace("\x1b[0m", NORMAL + color)


def ljust_colored(x: str, length: int, char: str = " ") -> str:
    remaining = length - len(strip(x))
    return x + ((remaining // len(char) + 1) * char)[:remaining]


def rjust_colored(x: str, length: int, char: str = " "):
    remaining = length - len(strip(x))
    return ((remaining // len(char) + 1) * char)[:remaining] + x



"""
Displays information about loaded kernel modules. This command retrieves the list of kernel modules from the `modules` symbol
and displays information about each module. It can filter modules by a substring of their names if provided.
"""

from __future__ import annotations

import argparse

from tabulate import tabulate

import pwndbg
import pwndbg.aglib.kernel.kmod
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.color.message as message
import pwndbg.commands

parser = argparse.ArgumentParser(description="Displays the loaded Linux kernel modules.")
parser.add_argument(
    "module_name", nargs="?", type=str, help="A module name substring to filter for"
)
parser.add_argument("-l", "--load", dest="path", type=str, help="the path of the module to load")


@pwndbg.commands.Command(parser, category=pwndbg.commands.CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.commands.OnlyWithKernelSymbols
def kmod(module_name: str | None = None, path: str | None = None) -> None:
    # Look up the address of the `modules` symbol, containing the head of the linked list of kernel modules
    modules_head = pwndbg.aglib.kernel.modules()
    if modules_head is None:
        print(
            "The modules symbol was not found. This may indicate that the symbol is not available in the current build."
        )
        return

    print(f"Kernel modules address found at {int(modules_head):#x}.\n")

    table = []
    headers = ["Address", "Name", "Size", "Used by"]
    krelease = pwndbg.aglib.kernel.krelease()
    if pwndbg.aglib.typeinfo.load("struct module") is not None:
        # Iterate through the linked list of modules using for_each_entry
        for module in pwndbg.aglib.kernel.kmod.module_list_with_typeinfo():
            name_addr = module["name"].address
            assert name_addr
            name = pwndbg.aglib.memory.string(int(name_addr)).decode("utf-8", errors="ignore")
            if not krelease or krelease >= (6, 4):
                addr = int(module["mem"][0]["base"])
                size = sum(
                    int(module["mem"][i]["size"])
                    for i in range(pwndbg.aglib.kernel.kmod.mod_mem_type.MOD_MEM_NUM_TYPES.value)
                )
            else:
                addr = int(module["core_layout"]["base"])
                size = int(module["core_layout"]["size"])
            uses = int(module["refcnt"]["counter"]) - 1

            # If module_name is provided, filter modules by name substring
            if not module_name or module_name in name:
                table.append([f"{addr:#x}", name, size, uses])
    else:
        cur = pwndbg.aglib.memory.read_pointer_width(int(modules_head))
        name_offset = pwndbg.aglib.kernel.kmod.module_name_offset()
        if not name_offset:
            print(message.warn("module->name offset not found"))
            return
        for cur in pwndbg.aglib.kernel.kmod.module_list():
            name = pwndbg.aglib.memory.string(cur + name_offset).decode()
            if not krelease or krelease >= (6, 4):
                mem_offset, module_memory_size, size_offset = (
                    pwndbg.aglib.kernel.kmod.module_mem_offset()
                )
                if mem_offset is None or module_memory_size is None or size_offset is None:
                    continue
                addr = pwndbg.aglib.memory.read_pointer_width(cur + mem_offset)
                size = 0
                for i in range(pwndbg.aglib.kernel.kmod.mod_mem_type.MOD_MEM_NUM_TYPES.value):
                    ptr = cur + mem_offset + module_memory_size * i
                    size += pwndbg.aglib.memory.u32(ptr + size_offset)
            else:
                addr_offset, size_offset = pwndbg.aglib.kernel.kmod.module_layout_offset()
                if addr_offset is None or size_offset is None:
                    continue
                addr = pwndbg.aglib.memory.read_pointer_width(cur + addr_offset)
                size = pwndbg.aglib.memory.u32(cur + size_offset)

            if not module_name or module_name in name:
                table.append([f"{addr:#x}", name, size, "-"])
    if path is not None:
        if len(table) == 1:
            addr = table[0][0]
            pwndbg.dbg.selected_inferior().add_symbol_file(path, addr)
            # FIXME: Reintroduce rizin/radare2 support here.
            # if pwndbg.config.decompiler == "radare2":
            #     pwndbg.radare2.r2cmd(["o", path, addr])
            # elif pwndbg.config.decompiler == "rizin":
            #     pwndbg.rizin.rzcmd(["o", path, addr])
        elif len(table) > 1:
            print(message.warn("Multiple modules detected with the given filter"))
        else:
            print(message.warn("No modules detected with the given filter."))
        return

    print(tabulate(table, headers=headers, tablefmt="simple"))



from __future__ import annotations

KNOWN_AIDS = {
    0: "AID_ROOT",
    1000: "AID_SYSTEM",
    1001: "AID_RADIO",
    1002: "AID_BLUETOOTH",
    1003: "AID_GRAPHICS",
    1004: "AID_INPUT",
    1005: "AID_AUDIO",
    1006: "AID_CAMERA",
    1007: "AID_LOG",
    1008: "AID_COMPASS",
    1009: "AID_MOUNT",
    1010: "AID_WIFI",
    1011: "AID_ADB",
    1012: "AID_INSTALL",
    1013: "AID_MEDIA",
    1014: "AID_DHCP",
    1015: "AID_SDCARD_RW",
    1016: "AID_VPN",
    1017: "AID_KEYSTORE",
    1018: "AID_USB",
    1019: "AID_DRM",
    1020: "AID_MDNSR",
    1021: "AID_GPS",
    1022: "AID_UNUSED1",
    1023: "AID_MEDIA_RW",
    1024: "AID_MTP",
    1025: "AID_UNUSED2",
    1026: "AID_DRMRPC",
    1027: "AID_NFC",
    1028: "AID_SDCARD_R",
    1029: "AID_CLAT",
    1030: "AID_LOOP_RADIO",
    1031: "AID_MEDIA_DRM",
    1032: "AID_PACKAGE_INFO",
    1033: "AID_SDCARD_PICS",
    1034: "AID_SDCARD_AV",
    1035: "AID_SDCARD_ALL",
    1036: "AID_LOGD",
    1037: "AID_SHARED_RELRO",
    1038: "AID_DBUS",
    1039: "AID_TLSDATE",
    1040: "AID_MEDIA_EX",
    1041: "AID_AUDIOSERVER",
    1042: "AID_METRICS_COLL",
    1043: "AID_METRICSD",
    1044: "AID_WEBSERV",
    1045: "AID_DEBUGGERD",
    1046: "AID_MEDIA_CODEC",
    1047: "AID_CAMERASERVER",
    1048: "AID_FIREWALL",
    1049: "AID_TRUNKS",
    1050: "AID_NVRAM",
    2001: "AID_CACHE",
    2002: "AID_DIAG",
    2900: "AID_OEM_RESERVED_START",
    2999: "AID_OEM_RESERVED_END",
    3001: "AID_NET_BT_ADMIN",
    3002: "AID_NET_BT",
    3003: "AID_INET",
    3004: "AID_NET_RAW",
    3005: "AID_NET_ADMIN",
    3006: "AID_NET_BW_STATS",
    3007: "AID_NET_BW_ACCT",
    3008: "AID_NET_BT_STACK",
    3009: "AID_READPROC",
    3010: "AID_WAKELOCK",
    5000: "AID_OEM_RESERVED_2_START",
    5999: "AID_OEM_RESERVED_2_END",
    9997: "AID_EVERYBODY",
    9998: "AID_MISC",
    9999: "AID_NOBODY",
    10000: "AID_APP",
    50000: "AID_SHARED_GID_START",
    59999: "AID_SHARED_GID_END",
    99000: "AID_ISOLATED_START",
    99999: "AID_ISOLATED_END",
    100000: "AID_USER",
}


def aid_name(uid: int) -> str:
    if uid in KNOWN_AIDS:
        return KNOWN_AIDS[uid]

    for closest in sorted(KNOWN_AIDS, reverse=True):
        if uid > closest:
            break
    else:
        return str(uid)

    return f"{KNOWN_AIDS[closest]}+{uid - closest}"



from __future__ import annotations

from typing import Any

import pwndbg
import pwndbg.commands
import pwndbg.lib.config as cfg


def pget(name: str) -> pwndbg.lib.config.Parameter | None:
    """
    Retrieves a parameter with a given name.
    """
    return pwndbg.config.params.get(name.replace("-", "_"))


def pset(param: pwndbg.lib.config.Parameter, value: str):
    """
    Parses and sets a Pwndbg configuration value.

    Raises `InvalidParse` if the value is not valid.
    """
    new_value = parse_value(param, value)

    param.value = new_value
    for trigger in pwndbg.config.triggers[param.name]:
        trigger()

    return True


class InvalidParse(Exception):
    pass


def parse_value(param: pwndbg.lib.config.Parameter, expression: str) -> Any:
    param_class = param.param_class
    if param_class == cfg.PARAM_BOOLEAN:
        if expression == "on":
            return True
        if expression == "off":
            return False
        raise InvalidParse("expected 'on' or 'off'")

    if param_class == cfg.PARAM_ZINTEGER:
        try:
            return int(expression, 0)
        except ValueError:
            raise InvalidParse("expected an integer value")
    elif param_class == cfg.PARAM_STRING:
        # We have to resolve any escape sequences that may present in the string
        # we received as our expression. Exactly which escape sequences should
        # be handled is not specified, by either us or GDB, which we base this
        # behavior on, even in the LLDB version.
        #
        # Ultimately, we use the native handling of escape sequences in the
        # "unicode_escape" decoder, and hope that it's good enough. Keep in mind
        # that what we're doing here is encoding the string to an ASCII string
        # with unicode escape sequences, and then decoding it as an ASCII string
        # with escape sequences. This allows us to retain any unicode originally
        # in the string, while at the same time resolving all escape sequences
        # in the original string.
        return expression.encode("ascii", "backslashreplace").decode("unicode_escape")
    elif param_class == cfg.PARAM_ZUINTEGER or param_class == cfg.PARAM_ZUINTEGER_UNLIMITED:
        try:
            value = int(expression, 0)
            if value < 0:
                raise InvalidParse("value must be greater than or equal to zero")
            return value
        except ValueError:
            raise InvalidParse("expected an integer value")
    elif param_class == cfg.PARAM_ENUM:
        if expression not in param.enum_sequence:
            names = ", ".join([f"'{name}'" for name in param.enum_sequence])
            raise InvalidParse(f"expected one of {names}")
        return expression
    elif param_class == cfg.PARAM_OPTIONAL_FILENAME:
        # We just hope the name is correct :)
        return expression
    elif param_class == cfg.PARAM_AUTO_BOOLEAN:
        if expression == "on":
            return True
        if expression == "off":
            return False
        if expression == "auto":
            return None
        raise InvalidParse("expected 'on', 'off', or 'auto'")
    elif param_class == cfg.PARAM_INTEGER:
        try:
            value = int(expression, 0)
            if value == 0:
                raise InvalidParse("value must not be zero")
            return value
        except ValueError:
            raise InvalidParse("expected an integer value")
    elif param_class == cfg.PARAM_UINTEGER:
        try:
            value = int(expression, 0)
            if value <= 0:
                raise InvalidParse("value must be greater than zero")
            return value
        except ValueError:
            raise InvalidParse("expected an integer value")

    raise NotImplementedError(f"Unknown parameter class {param_class}")



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from capstone6pwndbg import *  # noqa: F403
from capstone6pwndbg.aarch64 import *  # noqa: F403
from typing_extensions import override

import pwndbg.aglib
import pwndbg.aglib.disasm.arch
import pwndbg.lib.disasm.helpers as bit_math
from pwndbg.aglib.disasm.arch import register_assign
from pwndbg.aglib.disasm.instruction import ALL_JUMP_GROUPS
from pwndbg.aglib.disasm.instruction import EnhancedOperand
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.aglib.disasm.instruction import boolean_to_instruction_condition

# Emulator currently requires GDB, and we only use it here for type checking.
if TYPE_CHECKING:
    from pwndbg.emu.emulator import Emulator

# Negative size indicates signed read
# None indicates the read size depends on the target register
AARCH64_SINGLE_LOAD_INSTRUCTIONS: dict[int, int | None] = {
    AARCH64_INS_LDRB: 1,
    AARCH64_INS_ALIAS_LDRB: 1,
    AARCH64_INS_LDURB: 1,
    AARCH64_INS_ALIAS_LDURB: 1,
    AARCH64_INS_LDRSB: -1,
    AARCH64_INS_ALIAS_LDRSB: -1,
    AARCH64_INS_LDURSB: -1,
    AARCH64_INS_ALIAS_LDURSB: -1,
    AARCH64_INS_LDRH: 2,
    AARCH64_INS_ALIAS_LDRH: 2,
    AARCH64_INS_LDURH: 2,
    AARCH64_INS_ALIAS_LDURH: 2,
    AARCH64_INS_LDRSH: -2,
    AARCH64_INS_ALIAS_LDRSH: -2,
    AARCH64_INS_LDURSH: -2,
    AARCH64_INS_ALIAS_LDURSH: -2,
    AARCH64_INS_LDURSW: -4,
    AARCH64_INS_ALIAS_LDURSW: -4,
    AARCH64_INS_LDRSW: -4,
    AARCH64_INS_ALIAS_LDRSW: -4,
    AARCH64_INS_LDUR: None,
    AARCH64_INS_ALIAS_LDUR: None,
    AARCH64_INS_LDR: None,
    AARCH64_INS_ALIAS_LDR: None,
    AARCH64_INS_LDTRB: 1,
    AARCH64_INS_LDTRSB: -1,
    AARCH64_INS_LDTRH: 2,
    AARCH64_INS_LDTRSH: -2,
    AARCH64_INS_LDTRSW: -4,
    AARCH64_INS_LDTR: None,
    AARCH64_INS_ALIAS_LDTR: None,
    AARCH64_INS_LDXRB: 1,
    AARCH64_INS_LDXRH: 2,
    AARCH64_INS_LDXR: None,
    AARCH64_INS_LDARB: 1,
    AARCH64_INS_LDARH: 2,
    AARCH64_INS_LDAR: None,
}

# None indicates that the write size depends on the source register
AARCH64_SINGLE_STORE_INSTRUCTIONS: dict[int, int | None] = {
    AARCH64_INS_STRB: 1,
    AARCH64_INS_ALIAS_STRB: 1,
    AARCH64_INS_STURB: 1,
    AARCH64_INS_ALIAS_STURB: 1,
    AARCH64_INS_STRH: 2,
    AARCH64_INS_ALIAS_STRH: 2,
    AARCH64_INS_STURH: 2,
    AARCH64_INS_ALIAS_STURH: 2,
    AARCH64_INS_STUR: None,
    AARCH64_INS_ALIAS_STUR: None,
    AARCH64_INS_STR: None,
    AARCH64_INS_ALIAS_STR: None,
    # Store Register (unprivileged)
    AARCH64_INS_STTRB: 1,
    AARCH64_INS_STTRH: 2,
    AARCH64_INS_STTR: None,
    # Store-Release
    AARCH64_INS_STLRB: 1,
    AARCH64_INS_STLRH: 2,
    AARCH64_INS_STLR: None,
}

# The first operand of these instructions gets the status result of the operation
AARCH64_EXCLUSIVE_STORE_INSTRUCTIONS = {
    # Store Exclusive
    AARCH64_INS_STXRB: 1,
    AARCH64_INS_STXRH: 2,
    AARCH64_INS_STXR: None,
    # Store-Release Exclusive
    AARCH64_INS_STLXRB: 1,
    AARCH64_INS_STLXRH: 2,
    AARCH64_INS_STLXR: None,
}

CONDITIONAL_SELECT_INSTRUCTIONS = {
    AARCH64_INS_CSEL,
    AARCH64_INS_CSINC,
    AARCH64_INS_CSINV,
    AARCH64_INS_CSNEG,
    AARCH64_INS_ALIAS_CSET,
    AARCH64_INS_ALIAS_CSETM,
    AARCH64_INS_ALIAS_CINC,
    AARCH64_INS_ALIAS_CINV,
    AARCH64_INS_ALIAS_CNEG,
}

AARCH64_EMULATED_ANNOTATIONS = CONDITIONAL_SELECT_INSTRUCTIONS | {
    AARCH64_INS_SXTB,
    AARCH64_INS_SXTH,
    AARCH64_INS_SXTW,
    AARCH64_INS_UXTB,
    AARCH64_INS_UXTH,
    AARCH64_INS_UXTW,
    AARCH64_INS_RBIT,
    AARCH64_INS_CLS,
    AARCH64_INS_CLZ,
    AARCH64_INS_ALIAS_BFXIL,
    AARCH64_INS_ALIAS_UBFIZ,
    AARCH64_INS_UBFM,
    AARCH64_INS_ALIAS_UBFX,
    AARCH64_INS_ALIAS_SBFIZ,
    AARCH64_INS_SBFM,
    AARCH64_INS_ALIAS_SBFX,
    AARCH64_INS_ALIAS_BFI,
    AARCH64_INS_NEG,
    AARCH64_INS_ALIAS_NEGS,
    AARCH64_INS_REV,
    AARCH64_INS_BIC,
    AARCH64_INS_BICS,
}

AARCH64_CONSTANT_SHIFTS = {AARCH64_SFT_LSL, AARCH64_SFT_LSR, AARCH64_SFT_ASR, AARCH64_SFT_ROR}

# Parameters to each function: (value, shift_amt, bit_width)
AARCH64_BIT_SHIFT_MAP: dict[int, Callable[[int, int, int], int]] = {
    AARCH64_SFT_LSL: bit_math.logical_shift_left,
    AARCH64_SFT_LSR: bit_math.logical_shift_right,
    AARCH64_SFT_ASR: bit_math.arithmetic_shift_right,
    AARCH64_SFT_ROR: bit_math.rotate_right,
}


# These are "Extend" operations - https://devblogs.microsoft.com/oldnewthing/20220728-00/?p=106912
# They take in a number, extract a byte, halfword, or word,
# and perform a zero- or sign-extend operation.
AARCH64_EXTEND_MAP: dict[int, Callable[[int], int]] = {
    AARCH64_EXT_UXTB: lambda x: x & ((1 << 8) - 1),
    AARCH64_EXT_UXTH: lambda x: x & ((1 << 16) - 1),
    AARCH64_EXT_UXTW: lambda x: x & ((1 << 32) - 1),
    AARCH64_EXT_UXTX: lambda x: x,  # UXTX has no effect. It extracts 64-bits from a 64-bit register.
    AARCH64_EXT_SXTB: lambda x: bit_math.to_signed(x, 8),
    AARCH64_EXT_SXTH: lambda x: bit_math.to_signed(x, 16),
    AARCH64_EXT_SXTW: lambda x: bit_math.to_signed(x, 32),
    AARCH64_EXT_SXTX: lambda x: bit_math.to_signed(x, 64),
}

AARCH64_MATH_INSTRUCTIONS = {
    AARCH64_INS_ADD: "+",
    AARCH64_INS_ALIAS_ADD: "+",
    AARCH64_INS_ADDS: "+",
    AARCH64_INS_ALIAS_ADDS: "+",
    AARCH64_INS_SUB: "-",
    AARCH64_INS_ALIAS_SUB: "-",
    AARCH64_INS_SUBS: "-",
    AARCH64_INS_ALIAS_SUBS: "-",
    AARCH64_INS_AND: "&",
    AARCH64_INS_ALIAS_AND: "&",
    AARCH64_INS_ANDS: "&",
    AARCH64_INS_ALIAS_ANDS: "&",
    AARCH64_INS_ORR: "|",
    AARCH64_INS_ALIAS_ORR: "|",
    AARCH64_INS_EOR: "^",
    AARCH64_INS_ALIAS_EOR: "^",
    AARCH64_INS_UDIV: "/",
    AARCH64_INS_SDIV: "/",
    AARCH64_INS_SMULH: "*",
    AARCH64_INS_SMULL: "*",
    AARCH64_INS_ALIAS_SMULL: "*",
    AARCH64_INS_UMULH: "*",
    AARCH64_INS_UMULL: "*",
    AARCH64_INS_ALIAS_UMULL: "*",
    AARCH64_INS_MUL: "*",
    AARCH64_INS_ALIAS_MUL: "*",
}

AARCH64_SHIFT_INSTRUCTIONS = {
    AARCH64_INS_LSL: "<<",
    AARCH64_INS_ALIAS_LSL: "<<",
    AARCH64_INS_LSR: ">>",
    AARCH64_INS_ALIAS_LSR: ">>",
    AARCH64_INS_ASR: ">>s",
    AARCH64_INS_ALIAS_ASR: ">>s",
    AARCH64_INS_ROR: ">>r",
    AARCH64_INS_ALIAS_ROR: ">>r",
}


def resolve_condition(condition: int, cpsr: int) -> InstructionCondition:
    """
    Given a condition and the NZCV flag bits, determine when the condition is satisfied

    The condition is a Capstone constant
    """

    n = (cpsr >> 31) & 1
    z = (cpsr >> 30) & 1
    c = (cpsr >> 29) & 1
    v = (cpsr >> 28) & 1

    condition = {
        AArch64CC_EQ: z == 1,
        AArch64CC_NE: z == 0,
        AArch64CC_HS: c == 1,
        AArch64CC_LO: c == 0,
        AArch64CC_MI: n == 1,
        AArch64CC_PL: n == 0,
        AArch64CC_VS: v == 1,
        AArch64CC_VC: v == 0,
        AArch64CC_HI: c == 1 and z == 0,
        AArch64CC_LS: not (c == 1 and z == 0),
        AArch64CC_GE: n == v,
        AArch64CC_LT: n != v,
        AArch64CC_GT: z == 0 and n == v,
        AArch64CC_LE: not (z == 0 and n == v),
        AArch64CC_AL: True,
        AArch64CC_NV: True,
        AArch64CC_Invalid: True,
    }.get(condition, False)

    return InstructionCondition.TRUE if condition else InstructionCondition.FALSE


class AArch64DisassemblyAssistant(pwndbg.aglib.disasm.arch.DisassemblyAssistant):
    def __init__(self, architecture) -> None:
        super().__init__(architecture)

        self.annotation_handlers: dict[int, Callable[[PwndbgInstruction, Emulator], None]] = {
            # MOV
            AARCH64_INS_MOV: self._common_move_annotator,
            AARCH64_INS_ALIAS_MOV: self._common_move_annotator,
            # MOVZ
            AARCH64_INS_MOVZ: self._common_move_annotator,
            # MOV WITH KEEP
            AARCH64_INS_MOVK: self._common_generic_register_destination,
            # ADR
            AARCH64_INS_ADR: self._common_generic_register_destination,
            # ADRP
            AARCH64_INS_ADRP: self._handle_adrp,
            # CMP
            AARCH64_INS_ALIAS_CMP: self._common_cmp_annotator_builder("cpsr", "-"),
            # CMN
            AARCH64_INS_ALIAS_CMN: self._common_cmp_annotator_builder("cpsr", "+"),
            # TST (bitwise "and")
            AARCH64_INS_ALIAS_TST: self._common_cmp_annotator_builder("cpsr", "&"),
            # CCMP (conditional compare)
            AARCH64_INS_CCMP: self._common_cmp_annotator_builder("cpsr", ""),
            # CCMN
            AARCH64_INS_CCMN: self._common_cmp_annotator_builder("cpsr", ""),
        }

    @override
    def _set_annotation_string(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        # Dispatch to the correct handler
        if instruction.id in AARCH64_SINGLE_LOAD_INSTRUCTIONS:
            target_reg_size = self._register_width(instruction, instruction.operands[0]) // 8
            read_size = AARCH64_SINGLE_LOAD_INSTRUCTIONS[instruction.id] or target_reg_size

            self._common_load_annotator(
                instruction,
                emu,
                instruction.operands[1].before_value,
                abs(read_size),
                read_size < 0,
                target_reg_size,
                instruction.operands[0].str,
                instruction.operands[1].str,
            )
        elif instruction.id in AARCH64_SINGLE_STORE_INSTRUCTIONS:
            self._common_store_annotator(
                instruction,
                emu,
                instruction.operands[1].before_value,
                instruction.operands[0].before_value,
                AARCH64_SINGLE_STORE_INSTRUCTIONS[instruction.id],
                instruction.operands[1].str,
            )
        elif instruction.id in AARCH64_EXCLUSIVE_STORE_INSTRUCTIONS:
            self._common_store_annotator(
                instruction,
                emu,
                instruction.operands[-1].before_value,
                instruction.operands[-2].before_value,
                AARCH64_EXCLUSIVE_STORE_INSTRUCTIONS[instruction.id],
                instruction.operands[-1].str,
            )
        elif instruction.id in AARCH64_MATH_INSTRUCTIONS:
            self._common_binary_op_annotator(
                instruction,
                emu,
                instruction.operands[0],
                instruction.operands[-2].before_value,
                instruction.operands[-1].before_value,
                AARCH64_MATH_INSTRUCTIONS[instruction.id],
            )
        elif instruction.id in AARCH64_SHIFT_INSTRUCTIONS:
            # AArch64 encoding of shifts forces special attention: https://github.com/capstone-engine/capstone/issues/2631
            if len(instruction.operands) == 2:
                if instruction.operands[1].cs_op.shift.type in AARCH64_CONSTANT_SHIFTS:
                    self._common_binary_op_annotator(
                        instruction,
                        emu,
                        instruction.operands[0],
                        instruction.operands[1].before_value_no_modifiers,
                        instruction.operands[1].cs_op.shift.value,
                        AARCH64_SHIFT_INSTRUCTIONS[instruction.id],
                    )
                else:
                    self._common_generic_register_destination(instruction, emu)
        elif instruction.id in AARCH64_EMULATED_ANNOTATIONS:
            self._common_generic_register_destination(instruction, emu)
        else:
            self.annotation_handlers.get(instruction.id, lambda *a: None)(instruction, emu)

    def _handle_adrp(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        result_operand, right = instruction.operands
        if result_operand.str and right.before_value is not None:
            address = right.before_value & pwndbg.aglib.arch.ptrmask

            TELESCOPE_DEPTH = max(0, int(pwndbg.config.disasm_telescope_depth))

            addresses = self._telescope(address, TELESCOPE_DEPTH, instruction, emu)

            telescope = self._telescope_format_list(addresses, TELESCOPE_DEPTH, emu)

            instruction.annotation = register_assign(result_operand.str, telescope)

    @override
    def _prepare(
        self, instruction: PwndbgInstruction, emu: pwndbg.aglib.disasm.arch.Emulator
    ) -> None:
        if CS_GRP_INT in instruction.groups:
            # https://github.com/capstone-engine/capstone/issues/2630
            instruction.groups.remove(CS_GRP_CALL)

    @override
    def _condition(
        self, instruction: PwndbgInstruction, emu: Emulator
    ) -> pwndbg.aglib.disasm.arch.InstructionCondition:
        # In ARM64, only branches have the conditional code in the instruction,
        # as opposed to ARM32 which allows most instructions to be conditional
        if instruction.id == AARCH64_INS_B:
            # The B instruction can be made conditional by the condition codes
            if instruction.cs_insn.cc in (AArch64CC_Invalid, AArch64CC_AL):
                instruction.declare_is_unconditional_jump = True
            else:
                flags = super()._read_register_name(instruction, "cpsr", emu)
                if flags is not None:
                    return resolve_condition(instruction.cs_insn.cc, flags)

        elif instruction.id == AARCH64_INS_CBNZ:
            op_val = instruction.operands[0].before_value
            return boolean_to_instruction_condition(op_val is not None and op_val != 0)

        elif instruction.id == AARCH64_INS_CBZ:
            op_val = instruction.operands[0].before_value
            return boolean_to_instruction_condition(op_val is not None and op_val == 0)

        elif instruction.id == AARCH64_INS_TBNZ:
            op_val, bit = (
                instruction.operands[0].before_value,
                instruction.operands[1].before_value,
            )

            if op_val is not None and bit is not None:
                return boolean_to_instruction_condition(bool((op_val >> bit) & 1))

        elif instruction.id == AARCH64_INS_TBZ:
            op_val, bit = (
                instruction.operands[0].before_value,
                instruction.operands[1].before_value,
            )

            if op_val is not None and bit is not None:
                return boolean_to_instruction_condition(not ((op_val >> bit) & 1))
        elif instruction.id in CONDITIONAL_SELECT_INSTRUCTIONS:
            # Capstone places the condition to be satisfied in the `cc` field of the instruction
            # for all conditional select instructions
            flags = self._read_register_name(instruction, "cpsr", emu)

            if flags is not None:
                return resolve_condition(instruction.cs_insn.cc, flags)

        return super()._condition(instruction, emu)

    @override
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        if not bool(instruction.groups & ALL_JUMP_GROUPS):
            return None

        if len(instruction.operands) > 0:
            # For all AArch64 branches, the target is either an immediate or a register and is the last operand
            if (val := instruction.operands[-1].before_value) is not None:
                return val & pwndbg.aglib.arch.ptrmask
            return None
        if instruction.id in (AARCH64_INS_RET, AARCH64_INS_ALIAS_RET):
            # If this is a ret WITHOUT an operand, it means we should read from the LR/x30 register
            return super()._read_register_name(instruction, "lr", emu)

        return super()._resolve_target(instruction, emu)

    @override
    def _parse_memory(
        self, instruction: PwndbgInstruction, op: EnhancedOperand, emu: Emulator
    ) -> int | None:
        """
        Parse the `Arm64OpMem` Capstone object to determine the concrete memory address used.

        Three types of AArch64 memory operands:
        1. Register base with optional immediate offset
        Examples:
              ldrb   w3, [x2]
              str    x1, [x2, #0xb58]
              ldr x4,[x3], 4
        2. Register + another register with an optional shift
        Examples:
              ldrb   w1, [x9, x2]
              str x1, [x2, x0, lsl #3]
        3. Register + 32-bit register extended and shifted.
        The shift in this case is implicitly a LSL
        Examples:
              ldr x1, [x2, w22, UXTW #3]

        """

        target = 0

        # Handle case of LDR instructions memory operands that uses immediate as memory location
        # Example: ldr x1, 0x30
        if op.mem.base == 0:
            return op.mem.disp

        base = self._read_register(instruction, op.mem.base, emu)
        if base is None:
            return None
        target = base

        if not instruction.cs_insn.post_index:
            # Unlike in arm32, the sign of `disp` is encoded in the number already.
            # There is no separate ".subtracted" value in Capstone
            target += op.mem.disp

        # If there is an index register
        if op.mem.index != 0:
            index = self._read_register(instruction, op.mem.index, emu)
            if index is None:
                return None

            # Optionally apply an extend to the index register
            if op.cs_op.ext != 0:
                index = AARCH64_EXTEND_MAP[op.cs_op.ext](index)

            # Optionally apply shift to the index register
            # This handles shifts in the extend operation as well:
            # As in the case of `ldr x1, [x2, w22, UXTW #3]`,
            # Capstone will automatically make the shift a LSL and set the value to 3
            if op.cs_op.shift.type != 0:
                # The form of instructions with a shift always apply the shift to a 64-bit value
                index = AARCH64_BIT_SHIFT_MAP[op.cs_op.shift.type](index, op.cs_op.shift.value, 64)

            target += index

        return target

    def _register_width(self, instruction: PwndbgInstruction, op: EnhancedOperand) -> int:
        return 32 if instruction.cs_insn.reg_name(op.reg)[0] == "w" else 64

    @override
    def _parse_immediate(self, instruction: PwndbgInstruction, op: EnhancedOperand, emu: Emulator):
        """
        In AArch64, there can be an optional shift applied to constants, typically only a `LSL #12`

        Ex:
            cmp    x8, #1, lsl #12      (1 << 12)
        """
        target = op.imm
        if target is None:
            return None

        if op.cs_op.shift.type != 0:
            target = AARCH64_BIT_SHIFT_MAP[op.cs_op.shift.type](target, op.cs_op.shift.value, 64)

        return target

    @override
    def _parse_register(
        self, instruction: PwndbgInstruction, op: EnhancedOperand, emu: Emulator
    ) -> int | None:
        """
        Register operands can have optional extend and shift modifiers.

        Ex:
            cmp x5, x3, LSL #12         (x3 << 12)
            cmp x5, w3, SXTB 4          (Signed extend byte, then left shift 4)

        The extend operation is always applied first (if present), and then shifts take effect.
        """
        target = super()._parse_register(instruction, op, emu)
        if target is None:
            return None

        # We need this to retain the value of the un-shifted register in some annotations, such as shifts
        op.before_value_no_modifiers = target

        # The shift and sign-extend operations depend on the target bit width.
        # This is sometimes implicit in the target register size, which is always
        # the first operand.
        target_bit_width = (
            self._register_width(instruction, instruction.operands[0])
            if instruction.operands[0].type == CS_OP_REG
            else 64
        )

        if op.cs_op.ext != 0:
            target = AARCH64_EXTEND_MAP[op.cs_op.ext](target) & ((1 << target_bit_width) - 1)

        if op.cs_op.shift.type != 0:
            target = AARCH64_BIT_SHIFT_MAP.get(op.cs_op.shift.type, lambda *a: None)(
                target, op.cs_op.shift.value, target_bit_width
            )

            if target is not None:
                target = target & ((1 << target_bit_width) - 1)

        return target



from __future__ import annotations

import argparse
import errno

import pwndbg.aglib.errno
import pwndbg.commands
import pwndbg.libc
from pwndbg.commands import CommandCategory

# Manually add error code 0 for "OK"
errno.errorcode[0] = "OK"  # type: ignore[index]

parser = argparse.ArgumentParser(
    description="Converts errno (or argument) to its string representation."
)
parser.add_argument(
    "err",
    type=int,
    nargs="?",
    default=None,
    help="Errno; if not passed, it is retrieved from __errno_location",
)


@pwndbg.commands.Command(parser, command_name="errno", category=CommandCategory.LINUX)
@pwndbg.commands.OnlyWhenRunning
def errno_(err: int | None) -> None:
    if err is None:
        err, err_str = pwndbg.aglib.errno.get()
        if err_str != "":
            print(err_str)
            if pwndbg.libc.which() == pwndbg.libc.LibcType.UNKNOWN:
                print("Is the libc not loaded yet?")
            return

    msg = errno.errorcode.get(int(err), "Unknown error code")
    print(f"Errno {err}: {msg}")



from __future__ import annotations

import argparse

import pwndbg.aglib.kernel.nftables
import pwndbg.commands
from pwndbg.commands import CommandCategory


def parse_nft_family(s: str) -> int:
    val = pwndbg.aglib.kernel.nftables.nftables_table_family.get(s)
    if val is None:
        allowed = ",".join(set(pwndbg.aglib.kernel.nftables.nftables_table_family.keys()))
        raise argparse.ArgumentTypeError(f"Incorrect family '{s}', only '{allowed}'")
    return val


parser = argparse.ArgumentParser(
    description="Dump all nftables: tables, chains, rules, expressions"
)
parser.add_argument("nsid", type=int, nargs="?", help="Network Namespace ID")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def knft_dump(nsid: int | None = None):
    nft = pwndbg.aglib.kernel.nftables.Nftables.find(nsid=nsid)
    if nft is None:
        print("No netns found")
        return

    nft.print()


parser = argparse.ArgumentParser(
    description="Dump netfliter tables from a specific network namespace"
)
parser.add_argument("--nsid", "-n", type=int, help="Network Namespace ID")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def knft_list_tables(nsid: int | None = None):
    nft = pwndbg.aglib.kernel.nftables.Nftables.find(nsid=nsid)
    if nft is None:
        print("No netns found")
        return

    for table in nft.iter_tables():
        table.print(print_nested=False)


parser = argparse.ArgumentParser(description="Dump netfilter chains form a specific table")
parser.add_argument("--nsid", "-n", type=int, help="Network Namespace ID")
parser.add_argument(
    "table_family",
    nargs="?",
    type=parse_nft_family,
    help="Netfilter table family (inet, ip, ip6, netdev, bridge, arp)",
)
parser.add_argument("table_name", nargs="?", type=str, help="Table name")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def knft_list_chains(
    table_family: int | None = None, table_name: str | None = None, nsid: int | None = None
):
    is_any = False
    for nft in pwndbg.aglib.kernel.nftables.Table.find(
        nsid=nsid, table_family=table_family, table_name=table_name
    ):
        is_any = True
        for chain in nft.iter_chains():
            chain.print(print_nested=False)

    if not is_any:
        print("No nftables table found")


parser = argparse.ArgumentParser(description="Dump netfilter rules form a specific chain")
parser.add_argument("--nsid", "-n", type=int, help="Network Namespace ID")
parser.add_argument(
    "table_family",
    nargs="?",
    type=parse_nft_family,
    help="Table family, eg: inet, ip, ip6, netdev, bridge, arp",
)
parser.add_argument("table_name", nargs="?", type=str, help="Table name")
parser.add_argument("chain_name", nargs="?", type=str, help="Chain name")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def knft_list_rules(
    table_family: int | None = None,
    table_name: str | None = None,
    chain_name: str | None = None,
    nsid: int | None = None,
):
    is_any = False
    for nft in pwndbg.aglib.kernel.nftables.Chain.find(
        nsid=nsid, table_family=table_family, table_name=table_name, chain_name=chain_name
    ):
        is_any = True
        for rule in nft.iter_rules():
            rule.print(print_nested=False)
    if not is_any:
        print("No nftables chain found")


parser = argparse.ArgumentParser(description="Dump only expressions from specific rule")
parser.add_argument("--nsid", "-n", type=int, help="Network Namespace ID")
parser.add_argument(
    "table_family",
    nargs="?",
    type=parse_nft_family,
    help="Table family, eg: inet, ip, ip6, netdev, bridge, arp",
)
parser.add_argument("table_name", nargs="?", type=str, help="Table name")
parser.add_argument("chain_name", nargs="?", type=str, help="Chain name")
parser.add_argument("rule_id", nargs="?", type=int, help="Rule Handle ID")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def knft_list_exprs(
    table_family: int | None = None,
    table_name: str | None = None,
    chain_name: str | None = None,
    rule_id: int | None = None,
    nsid: int | None = None,
):
    is_any = False
    for nft in pwndbg.aglib.kernel.nftables.Rule.find(
        nsid=nsid,
        table_family=table_family,
        table_name=table_name,
        chain_name=chain_name,
        rule_id=rule_id,
    ):
        is_any = True
        for expr in nft.iter_exprs():
            expr.print(print_nested=True)
    if not is_any:
        print("No nftables rule found")


parser = argparse.ArgumentParser(description="Dump netfilter sets from a specific table")
parser.add_argument("--nsid", "-n", type=int, help="Network Namespace ID")
parser.add_argument(
    "table_family",
    nargs="?",
    type=parse_nft_family,
    help="Table family, eg: inet, ip, ip6, netdev, bridge, arp",
)
parser.add_argument("table_name", nargs="?", type=str, help="Table name")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def knft_list_sets(
    table_family: int | None = None, table_name: str | None = None, nsid: int | None = None
):
    is_any = False
    for nft in pwndbg.aglib.kernel.nftables.Table.find(
        nsid=nsid, table_family=table_family, table_name=table_name
    ):
        is_any = True
        for nft_set in nft.iter_sets():
            nft_set.print(print_nested=True)
    if not is_any:
        print("No nftables table found")


parser = argparse.ArgumentParser(description="Dump netfilter objects from a specific table")
parser.add_argument("--nsid", "-n", type=int, help="Network Namespace ID")
parser.add_argument(
    "table_family",
    nargs="?",
    type=parse_nft_family,
    help="Table family, eg: inet, ip, ip6, netdev, bridge, arp",
)
parser.add_argument("table_name", nargs="?", type=str, help="Table name")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def knft_list_objects(
    table_family: int | None = None, table_name: str | None = None, nsid: int | None = None
):
    is_any = False
    for nft in pwndbg.aglib.kernel.nftables.Table.find(
        nsid=nsid, table_family=table_family, table_name=table_name
    ):
        is_any = True
        for nft_object in nft.iter_objects():
            nft_object.print(print_nested=True)
    if not is_any:
        print("No nftables table found")


parser = argparse.ArgumentParser(description="Dump netfilter flowtables from a specific table")
parser.add_argument("--nsid", "-n", type=int, help="Network Namespace ID")
parser.add_argument(
    "table_family",
    nargs="?",
    type=parse_nft_family,
    help="Netfilter table family (inet, ip, ip6, netdev, bridge, arp)",
)
parser.add_argument("table_name", nargs="?", type=str, help="Table name")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelDebugInfo
@pwndbg.commands.OnlyWhenPagingEnabled
def knft_list_flowtables(
    table_family: int | None = None, table_name: str | None = None, nsid: int | None = None
):
    is_any = False
    for nft in pwndbg.aglib.kernel.nftables.Table.find(
        nsid=nsid, table_family=table_family, table_name=table_name
    ):
        is_any = True
        for flowtable in nft.iter_flowtables():
            flowtable.print(print_nested=True)
    if not is_any:
        print("No nftables table found")



from __future__ import annotations

import argparse

import gdb
from capstone6pwndbg import CS_GRP_JUMP

import pwndbg.aglib.disasm.disassembly
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.gdblib.bpoint
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.commands import CommandCategory


class BreakOnConditionalBranch(pwndbg.gdblib.bpoint.Breakpoint):
    """
    A breakpoint that only stops the inferior if a given branch is taken or not taken.
    """

    def __init__(self, instruction: PwndbgInstruction, taken: bool) -> None:
        super().__init__(f"*{instruction.address:#x}", type=gdb.BP_BREAKPOINT, internal=False)
        self.instruction = instruction
        self.taken = taken

    def should_stop(self):
        # We need to re-run the enhancement process, since now the PC == instruction.address,
        # where previously it was not. The enhancement process will figure out if all the conditions
        # this branch requires in order to be taken have been met.
        assistant = pwndbg.aglib.disasm.disassembly.get_disassembly_assistant_for_current_arch()
        assistant.enhance(self.instruction)
        condition_met = self.instruction.is_conditional_jump_taken

        return condition_met == self.taken


parser = argparse.ArgumentParser(description="Breaks on a branch if it is taken.")
parser.add_argument(
    "branch",
    type=str,
    help="The branch instruction to break on.",
)


@pwndbg.commands.Command(parser, command_name="break-if-taken", category=CommandCategory.BREAKPOINT)
@pwndbg.commands.OnlyWhenRunning
def break_if_taken(branch) -> None:
    install_breakpoint(branch, taken=True)


parser = argparse.ArgumentParser(description="Breaks on a branch if it is not taken.")
parser.add_argument(
    "branch",
    type=str,
    help="The branch instruction to break on.",
)


@pwndbg.commands.Command(
    parser, command_name="break-if-not-taken", category=CommandCategory.BREAKPOINT
)
@pwndbg.commands.OnlyWhenRunning
def break_if_not_taken(branch) -> None:
    install_breakpoint(branch, taken=False)


def install_breakpoint(branch, taken: bool) -> None:
    # Do our best to interpret branch as an address locspec. Untimately, though,
    # we're limited in what we can do from inside Python in that front.
    #
    # https://sourceware.org/gdb/onlinedocs/gdb/Address-Locations.html#Address-Locations
    address = None
    try:
        # Try to interpret branch as an address literal
        address = int(branch, 0)
    except ValueError:
        # That didn't work. Defer to GDB and see if it can make something out of
        # the address value we were given.
        try:
            value = gdb.parse_and_eval(branch)
            if value.address is None:
                print(message.warn(f"Value {branch} has no address, trying its value"))
                address = int(value)
            else:
                address = int(value.address)
        except gdb.error as e:
            # No such luck. Report to the user and quit.
            print(message.error(f"Could not resolve branch location {branch}: {e}"))
            return

    # We should've picked something by now, or errored out.
    instruction = pwndbg.aglib.disasm.disassembly.one(address)
    if instruction is None:
        print(message.error(f"Could not decode instruction at address {address:#x}"))
        return
    if CS_GRP_JUMP not in instruction.groups:
        print(
            message.error(
                f"Instruction '{instruction.mnemonic} {instruction.op_str}' at address {address:#x} is not a branch"
            )
        )
        return

    # Not all architectures have assistants we can use for conditionals.
    if not pwndbg.aglib.disasm.disassembly.arch_has_disassembly_assistant():
        print(
            message.error(
                "The current architecture is not supported for breaking on conditional branches"
            )
        )
        return

    # Install the breakpoint.
    BreakOnConditionalBranch(instruction, taken)



from __future__ import annotations

import argparse

import pwndbg.commands
import pwndbg.gdblib.ptmalloc2_tracking
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""Manages the heap tracker.

The heap tracker is a module that tracks usage of the GLibc heap and looks for
user errors such as double frees and use after frees.

Currently, the following errors can be detected:
    - Use After Free
""",
)

subparsers = parser.add_subparsers(
    required=True, description="Used to enable, disable and query information about the tracker"
)

# Subcommand that enables the tracker.
enable = subparsers.add_parser(
    "enable",
    help="Enable heap tracking",
    description="Enable heap tracking.",
)
enable.add_argument(
    "-b",
    "--hardware-breakpoints",
    dest="use_hardware_breakpoints",
    type=bool,
    default=False,
    help="Force the tracker to use hardware breakpoints.",
)
enable.set_defaults(mode="enable")

# Subcommand that disables the tracker.
disable = subparsers.add_parser(
    "disable", help="Disable heap tracking", description="Disable heap tracking."
)
disable.set_defaults(mode="disable")

# Subcommand that produces a report.
toggle_break = subparsers.add_parser(
    "toggle-break",
    help="Toggles whether possible UAF conditions will pause execution",
    description="Toggles whether possible UAF conditions will pause execution.",
)
toggle_break.set_defaults(mode="toggle-break")


@pwndbg.commands.Command(parser, category=CommandCategory.LINUX, command_name="track-heap")
@pwndbg.commands.OnlyWhenRunning
def track_heap(mode=None, use_hardware_breakpoints=False):
    if mode == "enable":
        # Enable the tracker.
        pwndbg.gdblib.ptmalloc2_tracking.install()
    elif mode == "disable":
        # Disable the tracker.
        pwndbg.gdblib.ptmalloc2_tracking.uninstall()
    elif mode == "toggle-break":
        # Delegate to the report function.
        pwndbg.gdblib.ptmalloc2_tracking.stop_on_error = (
            not pwndbg.gdblib.ptmalloc2_tracking.stop_on_error
        )
        if pwndbg.gdblib.ptmalloc2_tracking.stop_on_error:
            print("The program will stop when the heap tracker detects an error")
        else:
            print("The heap tracker will only print a message when it detects an error")
    else:
        raise AssertionError(f"track-heap must never have invalid mode '{mode}'. this is a bug")



from __future__ import annotations

import pwndbg.lib.config

config: pwndbg.lib.config.Config = pwndbg.lib.config.Config()
"""
The global configuration object.
"""

import pwndbg.color
import pwndbg.exception
import pwndbg.lib.version
import pwndbg.ui
from pwndbg.dbg_mod import dbg as dbg

__version__ = pwndbg.lib.version.__version__
"""Pwndbg version."""
version = __version__
"""Pwndbg version."""

# Don't know where else to put this xd
config.add_param("dev-debug-events", False, "display internal event debugging info")



from __future__ import annotations

import argparse
import re
from typing import Any

import pwndbg.aglib.proc
import pwndbg.aglib.symbol
import pwndbg.aglib.vmmap
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.gdblib.got
from pwndbg.commands import CommandCategory


def columns(rows, colors=None) -> None:
    """
    Print data formatted into distinct columns.
    """
    if len(rows) == 0:
        # Nothing to print.
        return

    col_max = [0] * len(rows[0])
    for i in range(len(rows)):
        if len(rows[i]) == 0:
            continue
        for j in range(len(col_max)):
            if len(rows[i][j]) > col_max[j]:
                col_max[j] = len(rows[i][j])

    for i in range(len(rows)):
        if len(rows[i]) == 0:
            print()
            continue

        for j in range(len(col_max)):
            color = colors[j] if colors is not None else lambda x: x
            print(f"{color(rows[i][j].ljust(col_max[j]))} ", end="")
        print()


parser = argparse.ArgumentParser(
    description="Controls GOT tracking",
)
subparsers = parser.add_subparsers(
    required=True, description="Used to disable and query information about the tracker"
)

# Subcommand that enables the tracker.
enable = subparsers.add_parser(
    "enable",
    help="Enable GOT parsing",
    description="Enable GOT parsing.",
)
enable.set_defaults(mode="enable")

# Subcommand that disables the tracker.
disable = subparsers.add_parser(
    "disable", help="Disable GOT tracking", description="Disable GOT tracking."
)
disable.set_defaults(mode="disable")

# Subcommand that produces a report.
report = subparsers.add_parser(
    "info",
    help="Give an overview of the GOT tracker",
    description="Give an overview of the GOT tracker.",
)
report.add_argument(
    "-s",
    "--so-name",
    type=str,
    dest="soname",
    default=".*",
    help="Selects objects whose names matche the given expression",
)
report.add_argument(
    "-w",
    "--writable",
    dest="writable",
    action="store_true",
    help="Only show functions whose GOT entry is in a writable region of memory",
)
report.add_argument(
    "-f",
    "--function-name",
    type=str,
    dest="fnname",
    default=".*",
    help="Selects functions whose names match the given expression",
)
report.set_defaults(mode="report")

# Subcommand that queries for information about a specific tracker.
status = subparsers.add_parser(
    "query",
    help="Queries detailed tracking information about a single entry in the GOT",
    description="Queries detailed tracking information about a single entry in the GOT.",
)
status.add_argument(
    "address",
    type=str,
    help="The address of the GOT entry being tracked",
)
status.set_defaults(mode="status")


@pwndbg.commands.Command(parser, category=CommandCategory.LINUX, command_name="track-got")
@pwndbg.commands.OnlyWhenRunning
def track_got(mode=None, soname=None, writable=False, fnname=None, address=None):
    if mode == "enable":
        # Enable the tracker.
        if pwndbg.gdblib.got.GOT_TRACKING:
            print("GOT tracking is already enabled. Did you mean to use a flag?")
            return
        pwndbg.gdblib.got.enable_got_call_tracking()
    elif mode == "disable":
        # Disable the tracker.
        if not pwndbg.gdblib.got.GOT_TRACKING:
            print("GOT tracking is already disabled. Did you mean to enable it with `track-got`?")
            return
        pwndbg.gdblib.got.disable_got_call_tracking()
    elif mode == "report":
        # Delegate to the report function.
        got_report(soname=soname, writable=writable, fnname=fnname)
    elif mode == "status":
        # Delegate to the status function.
        got_tracking_status(address=address)
    else:
        raise AssertionError(f"track-got must never have invalid mode '{mode}'. this is a bug")


def got_report(soname=".*", writable=False, fnname=".*") -> None:
    """
    Prints out a report of the current status of the GOT tracker.
    """
    if not pwndbg.gdblib.got.GOT_TRACKING:
        print(message.error("GOT call tracking is not enabled"))
        return

    soname = re.compile(soname)
    fnname = re.compile(fnname)

    print(
        f"Showing {'writable' if writable else 'all'} GOT function entries and how many times they were called."
    )
    print()

    per_object: dict[Any, Any] = {}
    for _, (tracker, patcher) in pwndbg.gdblib.got.all_tracked_entries():
        objname = tracker.link_map_entry.name()
        if objname == b"":
            objname = pwndbg.aglib.proc.exe()
        else:
            objname = pwndbg.gdblib.got.display_name(objname)

        # Filter out objects we're not interested in.
        if soname.match(objname) is None:
            continue

        if objname not in per_object:
            per_object[objname] = []
        per_object[objname].append((tracker, patcher))

    for objname, trackers in per_object.items():
        print(f"Calls from {objname}:")
        rows = [["Address in GOT", "Function Address", "Symbol", "Call Count"]]
        for tracker, patcher in trackers:
            # If requested, filter out entries that are not in a writable
            # portion of memory.
            if writable and not pwndbg.aglib.vmmap.find(patcher.entry).write:
                continue

            dynamic = tracker.dynamic_section
            sym_index = tracker.relocation_fn(tracker.relocation_index, "r_sym")
            sym_name = dynamic.symtab_read(sym_index, "st_name")
            sym_name = pwndbg.gdblib.got.display_name(dynamic.string(sym_name))

            # Filter out symbols we're not interested in.
            if fnname.match(sym_name) is None:
                continue

            if sym_name == "":
                sym_name = "<Empty>"

            hits = tracker.total_hits
            hits = f"{hits} hit{'s' if hits != 1 else ''}"
            rows.append([f"{patcher.entry:#x}", f"{tracker.target:#x}", sym_name, hits])
        columns(rows)
        print()


def got_tracking_status(address) -> None:
    """
    Prints out information about a single GOT tracking entry.
    """
    if not pwndbg.gdblib.got.GOT_TRACKING:
        print(message.error("GOT call tracking is not enabled"))
        return

    try:
        address = int(address, 0)
    except ValueError as e:
        print(message.error(f"Invalid address {address}: {e}"))
        return

    result = pwndbg.gdblib.got.tracked_entry_by_address(address)
    if result is None:
        print(message.error(f"No entry at address {address:#x}"))
        print("Hint: This command expects the address of the entry in the GOT. So, consider")
        print("using the address from the 'Address in GOT' column of the `track-got info`")
        print("command.")
        return

    tracker, _ = result

    dynamic = tracker.dynamic_section
    sym_index = tracker.relocation_fn(tracker.relocation_index, "r_sym")
    raw_sym_name = dynamic.symtab_read(sym_index, "st_name")
    sym_name = pwndbg.gdblib.got.display_name(dynamic.string(raw_sym_name))

    if sym_name == "":
        sym_name = "<Empty>"

    objname = tracker.link_map_entry.name()
    if objname == b"":
        objname = pwndbg.aglib.proc.exe()
    else:
        objname = pwndbg.gdblib.got.display_name(objname)

    print(f"Tracking details for {sym_name}")
    print()
    print(f"Dynamic object name: {objname}")
    print(f"Jump target address: {tracker.target:#x}")
    print(f"GOT entry address:   {address:#x}")
    print(f"Relocation index:    {tracker.relocation_index}")
    print(f"Symbol index:        {sym_index}")
    print()

    callers = sorted(tracker.hits.items(), key=lambda x: x[1])
    for stack, hits in callers:
        print(f"Called {hits} times from stack:")
        for entry in stack:
            print(f"    - {entry:#x} ", end="")
            symname = pwndbg.aglib.symbol.resolve_addr(entry)
            if symname != "":
                print(f"<{symname}>", end="")
            print()
        print()



"""
Utilities for profiling pwndbg.
"""

from __future__ import annotations

import argparse

import pwndbg.commands
import pwndbg.profiling
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""
Utilities for profiling Pwndbg.

Check out the `./profiling` folder for other useful utilities.
Use `./profiling/print_stats.py` to generate a report from a `.pstats` file.
"""
)
subparsers = parser.add_subparsers(dest="command")
parser_start = subparsers.add_parser("start", description="Starts the profiler.")
parser_stop = subparsers.add_parser("stop", description="Stop the profiler and save the results.")
parser_stop.add_argument(
    "--file", type=str, default="pwndbg.pstats", help="Output file for profile data."
)


@pwndbg.commands.Command(parser, category=CommandCategory.PWNDBG)
def profiler(command, file="pwndbg.pstats") -> None:
    if command == "start":
        print("Starting profiler.")
        pwndbg.profiling.profiler.start()
    elif command == "stop":
        print("Stopped profiler. Wrote results to", file)
        print(
            "To analyze the results, use the ./profiling/print_stats.py script in the pwndbg repo."
        )
        pwndbg.profiling.profiler.stop(file)
